"""
This module contains the parser that converts the string representation
of a sequence of digits into the corresponding sequence of digits. These digits may
be in the form of english cardinal representations of numbers, along with some
homophones. The digits can be hyphenated or unhyphenated from twenty through
ninety-nine. The unhyphenated numbers get joined automatically. The use of
unhyphenated numbers introduces ambiguity. For example, "sixty five thousand" could
be parsed as "605000" or "65000". Our parser will output the latter. However, this
can be an issue with values such as "sixty five thousand one" which parses as "650001".
This limitation will most likely be acceptable for most multi-digit use cases such as
telephone numbers, social security numbers, etc.
"""
from typing import Any, Dict, Union

from spokestack.nlu.parsers import DIGIT_SPLIT_RE, maps


def parse(metadata: Dict[str, Any], raw_value: str) -> str:
    """Digit Parser

    Args:
        metadata (Dict[str, Any]): digit slot metadata
        raw_value (str): value tagged by the model

    Returns:
        (str): string parsed digits
    """
    count = metadata.get("count")
    normalized = raw_value.lower()
    tokens = DIGIT_SPLIT_RE.split(normalized)
    values = []
    for i, token in enumerate(tokens):
        next_token = None
        if i < len(tokens) - 1:
            next_token = tokens[i + 1]
        value = _parse_single(token, next_token)
        values.append(value)

    combined = "".join(values)
    if count:
        if len(combined) != count:
            return ""
    return combined


def _parse_single(token: str, next_token: Union[str, None]) -> str:
    if token in maps.ENG_ZERO:
        return str(maps.ENG_ZERO[token])
    elif token in maps.ENG_MOD10:
        return str(maps.ENG_MOD10[token])
    elif token in maps.ENG_MOD20:
        return str(maps.ENG_MOD20[token])
    elif token in maps.ENG_DIV10 and next_token in maps.ENG_MOD10:
        return str(maps.ENG_DIV10[token])
    elif token in maps.ENG_DIV10:
        return str(maps.ENG_DIV10[token] * 10)
    elif token in maps.ENG_EXP10:
        exponent = maps.ENG_EXP10[token]
        return "0" * exponent
    else:
        try:
            return str(int(token))
        except ValueError:
            return ""



"""
This module contains the class for webrtc automatic noise suppression
"""
import numpy as np

from spokestack.context import SpeechContext
from spokestack.extensions.webrtc.nsx import WebRtcNsx

POLICY_MILD = 0
POLICY_MEDIUM = 1
POLICY_AGGRESSIVE = 2
POLICY_VERY_AGGRESSIVE = 3


class AutomaticNoiseSuppression:
    """WebRTC Automatic Noise Suppression

    Args:
        sample_rate (int): audio sample rate. (Hz)
        policy (int): aggressiveness of the noise suppression:
                      - POLICY_MILD: mild suppression (6dB)
                      - POLICY_MEDIUM: medium suppression (10dB)
                      - POLICY_AGGRESSIVE: aggresive suppression (15dB)
                      - POLICY_VERY_AGGRESSIVE: very aggressive suppression
    """

    def __init__(
        self, sample_rate: int = 16000, policy: int = POLICY_MEDIUM, **kwargs: str
    ) -> None:

        # validate sample rate
        if sample_rate not in {8000, 16000, 32000}:
            raise ValueError("sample_rate")

        self._nsx = WebRtcNsx(sample_rate=sample_rate, policy=policy)
        self._frame_width = sample_rate * 10 // 1000

    def __call__(self, context: SpeechContext, frame: np.array) -> None:
        """Main Entry Point

        Args:
            context (SpeechContext): State based information that needs to be shared
            between pieces of the pipeline
            frame (np.array): PCM-16 Audio
        """
        frame_size = self._frame_width
        # validate frame size
        if len(frame) % frame_size != 0:
            raise ValueError("invalid_frame_size")
        # validate dtype
        if not np.issubdtype(frame.dtype, np.signedinteger):
            raise TypeError("invalid_dtype")

        for i in range(len(frame) // frame_size):
            self._nsx(frame[i : frame_size + i])

    def close(self) -> None:
        """method for pipeline compliance"""
        pass

    def reset(self) -> None:
        """method for pipeline compliance"""
        pass



"""
This module manages the timeout for speech pipeline activation.
"""
from typing import Any, Union

import numpy as np

from spokestack.context import SpeechContext


class ActivationTimeout:
    """Speech pipeline activation timeout

    Args:
        frame_width (int): frame width of the audio (ms)
        min_active (int): the minimum length of an activation (ms)
        max_active (int): the maximum length of an activation (ms)
    """

    def __init__(
        self,
        frame_width: int = 20,
        min_active: int = 500,
        max_active: int = 5000,
        **kwargs: Any
    ) -> None:

        self._min_active = min_active / frame_width
        self._max_active = max_active / frame_width
        self._is_speech = False
        self._active_length = 0

    def __call__(
        self, context: SpeechContext, frame: Union[np.ndarray, None] = None
    ) -> None:
        """Main entry point that manages timeout

        Args:
            context (SpeechContext): the current state of the pipeline

        """
        vad_fall = self._is_speech and not context.is_speech
        self._is_speech = context.is_speech
        if context.is_active:
            self._active_length += 1
            if self._active_length > self._min_active:
                if vad_fall or self._active_length > self._max_active:
                    self.deactivate(context)

    def deactivate(self, context: SpeechContext) -> None:
        """ Deactivates the speech pipeline """
        self.reset()
        context.is_active = False

    def reset(self) -> None:
        """ Resets active length """
        self.close()

    def close(self) -> None:
        """ Sets active length to zero """
        self._active_length = 0



"""
This module contains the google asr speech recognizer
"""
import logging
from queue import Queue
from threading import Thread
from typing import Any, Generator, Union

import numpy as np
from google.cloud import speech
from google.oauth2 import service_account

from spokestack.context import SpeechContext

_LOG = logging.getLogger(__name__)


class GoogleSpeechRecognizer:
    """Transforms speech into text using Google's ASR.

    Args:
        language (str): The language of given audio as a
                        [BCP-47](https://www.rfc-editor.org/rfc/bcp/bcp47.txt)
                        language tag. Example: "en-US"
        credentials (Union[None, str, dict]): Dictionary of Google API credentials
                                              or path to credentials. if set to None
                                              credentials will be pulled from the
                                              environment variable:
                                              GOOGLE_APPLICATION_CREDENTIALS
        sample_rate (int): sample rate of the input audio (Hz)
        **kwargs (optional): additional keyword arguments
    """

    def __init__(
        self,
        language: str,
        credentials: Union[None, str, dict] = None,
        sample_rate: int = 16000,
        **kwargs: Any,
    ) -> None:
        if credentials:
            if isinstance(credentials, str):
                credentials = service_account.Credentials.from_service_account_file(
                    credentials
                )
            elif isinstance(credentials, dict):
                credentials = service_account.Credentials.from_service_account_info(
                    credentials
                )
            else:
                raise ValueError(
                    "Invalid Credentials: Only dict, str, or None accepted"
                )

        self._client = speech.SpeechClient(credentials=credentials)
        self._config = speech.StreamingRecognitionConfig(
            config=speech.RecognitionConfig(
                encoding=speech.RecognitionConfig.AudioEncoding.LINEAR16,
                sample_rate_hertz=sample_rate,
                language_code=language,
                enable_automatic_punctuation=True,
            ),
            interim_results=True,
        )
        self._queue: Queue = Queue()
        self._thread: Any = None

    def __call__(self, context: SpeechContext, frame: np.ndarray) -> None:
        """Main entry point.

        Args:
            context (SpeechContext): current state of the speech pipeline
            frame (np.ndarray): numpy array of PCM-16 audio.

        Returns: None

        """
        if self._thread is None and context.is_active:
            self._begin(context)
        if self._thread is not None and not context.is_active:
            self._commit()
        if context.is_active:
            self._send(frame)

    def _begin(self, context: SpeechContext) -> None:
        self._thread = Thread(
            target=self._receive,
            args=(context,),
        )
        self._thread.start()

    def _receive(self, context: SpeechContext) -> None:
        for response in self._client.streaming_recognize(self._config, self._drain()):
            for result in response.results[:1]:
                for alternative in result.alternatives[:1]:
                    context.transcript = alternative.transcript
                    context.confidence = alternative.confidence
                    if context.transcript:
                        context.event("partial_recognize")

                if result.is_final:
                    if context.transcript:
                        context.event("recognize")
                        _LOG.debug("recognize event")
                    else:
                        context.event("timeout")
                        _LOG.debug("timeout event")

    def _drain(self) -> Generator:
        while data := self._queue.get():
            yield data

    def _commit(self) -> None:
        self._queue.put(None)
        self._thread.join()
        self._thread = None

    def _send(self, frame: np.ndarray) -> None:
        self._queue.put(speech.StreamingRecognizeRequest(audio_content=frame.tobytes()))

    def reset(self) -> None:
        """ resets recognizer """
        if self._thread:
            self._queue.put(None)
            self._thread.join()
            self._thread = None

    def close(self) -> None:
        """ closes recognizer """
        self._client = None



"""
This module contains a context class to manage
state between members of the processing pipeline
"""
import logging
from typing import Callable


_LOG = logging.getLogger(__name__)


class SpeechContext:
    """ Class for managing context of the speech pipeline. """

    def __init__(self) -> None:
        self._is_speech: bool = False
        self._is_active: bool = False
        self._transcript: str = ""
        self._confidence: float = 0.0
        self._handlers: dict = {}

    def add_handler(self, name: str, function: Callable) -> None:
        """Adds a handler to the context

        Args:
            name (str): The name of the event handler
            function (Callable): event handler function

        """
        self._handlers[name] = function

    def event(self, name: str) -> None:
        """Calls the event handler

        Args:
            name (str): The name of the event handler

        """
        handler = self._handlers.get(name)
        if handler:
            handler(self)

    @property
    def is_speech(self) -> bool:
        """This property is to manage if speech is present in the current state
        or not.

        Returns:
            bool: 'True' if is_speech set to 'True', 'False' otherwise
        """
        return self._is_speech

    @is_speech.setter
    def is_speech(self, value: bool) -> None:
        """This method is the setter for the is_speech property.

        Args:
            value (bool): sets is_speech to passed argument
        """
        self._is_speech = value

    @property
    def is_active(self) -> bool:
        """This property manages activity of the context.

        Returns:
            bool: 'True' if context is active, 'False' otherwise.
        """
        return self._is_active

    @is_active.setter
    def is_active(self, value: bool) -> None:
        """This method sets the is_active property.

        Args:
            value (bool): Boolean to set context activity
        """
        is_active = self._is_active
        self._is_active = value
        if value and not is_active:
            self.event("activate")
            _LOG.info("activate event")
        elif not value and is_active:
            self.event("deactivate")
            _LOG.info("deactivate event")

    @property
    def transcript(self) -> str:
        """This property is the text representation of the audio buffer

        Returns:
            str: the value of the transcript property
        """
        return self._transcript

    @transcript.setter
    def transcript(self, value: str) -> None:
        """This method sets the transcript from a given string.

        Args:
            value (str): The text representation of speech input
        """
        self._transcript = value

    @property
    def confidence(self) -> float:
        """This property contains the confidence of a classification result.

        Returns:
            float: model confidence of classification
        """
        return self._confidence

    @confidence.setter
    def confidence(self, value: float) -> None:
        """This method sets the confidence property.

        Args:
            value (float): model confidence in classification result
        """
        self._confidence = value

    def reset(self) -> None:
        """Resets the context state"""
        self.is_speech = False
        self.is_active = False
        self.transcript = ""
        self.confidence = 0.0



"""
This module contains the Spokestack KeywordRecognizer which identifies multiple keywords
from an audio stream.
"""
import os
from typing import Any, List

import numpy as np

from spokestack.context import SpeechContext
from spokestack.models.tensorflow import TFLiteModel
from spokestack.ring_buffer import RingBuffer


class KeywordRecognizer:
    """Recognizes keywords in an audio stream.

    Args:
        classes (List[str]): Keyword labels
        pre_emphasis (float): The value of the pre-emphasis filter
        sample_rate (int): The number of audio samples per second of audio (kHz)
        fft_window_type (str): The type of fft window. (only support for hann)
        fft_hop_length (int): Audio sliding window for STFT calculation (ms)
        model_dir (str): Path to the directory containing .tflite models
        posterior_threshold (float): Probability threshold for detection
    """

    def __init__(
        self,
        classes: List[str],
        pre_emphasis: float = 0.97,
        sample_rate: int = 16000,
        fft_window_type: str = "hann",
        fft_hop_length: int = 10,
        model_dir: str = "",
        posterior_threshold: float = 0.5,
        **kwargs: Any
    ) -> None:

        self.classes = classes
        self.pre_emphasis: float = pre_emphasis
        self.hop_length: int = int(fft_hop_length * sample_rate / 1000)

        if fft_window_type != "hann":
            raise ValueError("Invalid fft_window_type")

        self.filter_model: TFLiteModel = TFLiteModel(
            model_path=os.path.join(model_dir, "filter.tflite")
        )
        self.encode_model: TFLiteModel = TFLiteModel(
            model_path=os.path.join(model_dir, "encode.tflite")
        )
        self.detect_model: TFLiteModel = TFLiteModel(
            model_path=os.path.join(model_dir, "detect.tflite")
        )

        if len(classes) != self.detect_model.output_details[0]["shape"][-1]:
            raise ValueError("Invalid number of classes")

        # window size calculated based on fft
        # the filter inputs are (fft_size - 1) / 2
        # which makes the window size (post_fft_size - 1) * 2
        self._window_size = (self.filter_model.input_details[0]["shape"][-1] - 1) * 2
        self._fft_window = np.hanning(self._window_size)

        # retrieve the mel_length and mel_width based on the encoder model metadata
        # these allocate the buffer to the correct size
        self.mel_length: int = self.encode_model.input_details[0]["shape"][1]
        self.mel_width: int = self.encode_model.input_details[0]["shape"][-1]

        # initialize the first state input for autoregressive encoder model
        # retrieve the encode_length and encode_width from the model detect_model
        # metadata. We get the dimensions from the detect_model inputs because the
        # encode_model runs autoregressive and outputs a single encoded sample.
        # the detect_model input is a collection of these samples.
        self.state = np.zeros(self.encode_model.input_details[1]["shape"], np.float32)
        self.encode_length: int = self.detect_model.input_details[0]["shape"][1]
        self.encode_width: int = self.detect_model.input_details[0]["shape"][-1]

        self.sample_window: RingBuffer = RingBuffer(
            shape=[self._window_size],
        )
        self.frame_window: RingBuffer = RingBuffer(
            shape=[self.mel_length, self.mel_width]
        )
        self.encode_window: RingBuffer = RingBuffer(
            shape=[self.encode_length, self.encode_width]
        )

        # initialize the frame and encode windows with zeros
        # this minimizes the delay caused by filling the buffer
        self.frame_window.fill(0.0)
        self.encode_window.fill(-1.0)

        self._posterior_threshold: float = posterior_threshold
        self._prev_sample: float = 0.0
        self._is_active = False

    def __call__(self, context: SpeechContext, frame: np.ndarray) -> None:

        self._sample(context, frame)

        if not context.is_active and self._is_active:
            self._detect(context)

        self._is_active = context.is_active

    def _sample(self, context: SpeechContext, frame: np.ndarray) -> None:
        # convert the PCM-16 audio to float32 in (-1.0, 1.0)
        frame = frame.astype(np.float32) / (2 ** 15 - 1)
        frame = np.clip(frame, -1.0, 1.0)

        # pull out a single value from the frame and apply pre-emphasis
        # with the previous sample then cache the previous sample
        # to be use in the next iteration
        prev_sample = frame[-1]
        frame -= self.pre_emphasis * np.append(self._prev_sample, frame[:-1])
        self._prev_sample = prev_sample

        # fill the sample window to analyze speech containing samples
        # after each window fill the buffer advances by the hop length
        # to produce an overlapping window
        for sample in frame:
            self.sample_window.write(sample)
            if self.sample_window.is_full:
                if context.is_active:
                    self._analyze(context)
                self.sample_window.rewind().seek(self.hop_length)

    def _analyze(self, context: SpeechContext) -> None:
        # read the full contents of the sample window to calculate a single frame
        # of the STFT by applying the DFT to a real-valued input and
        # taking the magnitude of the complex DFT
        frame = self.sample_window.read_all()
        frame = np.fft.rfft(frame * self._fft_window, n=self._window_size)
        frame = np.abs(frame).astype(np.float32)

        # compute mel spectrogram
        self._filter(context, frame)

    def _filter(self, context: SpeechContext, frame: np.ndarray) -> None:
        # add the batch dimension and compute the mel spectrogram with filter model
        frame = np.expand_dims(frame, 0)
        frame = self.filter_model(frame)[0]

        # advance the window by 1 and write mel frame to the frame buffer
        self.frame_window.rewind().seek(1)
        self.frame_window.write(frame)

        # encode the mel spectrogram
        self._encode(context)

    def _encode(self, context: SpeechContext) -> None:
        # read the full contents of the frame window and add the batch dimension
        # run the encoder and save the output state for autoregression
        frame = self.frame_window.read_all()
        frame = np.expand_dims(frame, 0)
        frame, self.state = self.encode_model(frame, self.state)

        # accumulate encoded samples until size of detection window
        self.encode_window.rewind().seek(1)
        self.encode_window.write(frame)

    def _detect(self, context: SpeechContext) -> None:
        # read the full contents of the encode window and add the batch dimension
        # calculate a scalar likelihood that the frame contains a keyword
        # with the detect model
        frame = self.encode_window.read_all()
        frame = np.expand_dims(frame, 0)
        posterior = self.detect_model(frame)[0][0]
        class_index = np.argmax(posterior)
        confidence = posterior[class_index]

        if confidence >= self._posterior_threshold:
            context.transcript = self.classes[class_index]
            context.confidence = confidence
            context.event("recognize")
        else:
            context.event("timeout")
        self.reset()

    def reset(self) -> None:
        """ Resets the current KeywordDetector state """
        self.sample_window.reset()
        self.frame_window.reset().fill(0.0)
        self.encode_window.reset().fill(-1.0)
        self.state[:] = 0.0

    def close(self) -> None:
        """ Close interface for use in the SpeechPipeline """
        self.reset()



"""
This module contains the tests for the activation timeout class
"""
from spokestack.activation_timeout import ActivationTimeout
from spokestack.context import SpeechContext


def test_timeout_vad_fall():
    max_active = 500
    min_active = 20
    context = SpeechContext()
    timeout = ActivationTimeout(min_active=min_active, max_active=max_active)

    context.is_active = True
    context.is_speech = False

    timeout(context)
    context.is_speech = True

    timeout(context)
    assert context.is_active

    context.is_speech = False

    steps_before_timeout = (min_active // 20) + 2
    for _ in range(steps_before_timeout):
        timeout(context)
    assert not context.is_active

    timeout.close()


def test_max_active():
    max_active = 500
    min_active = 20
    context = SpeechContext()
    timeout = ActivationTimeout(min_active=min_active, max_active=max_active)

    context.is_active = True

    steps_before_timeout = (max_active // 20) + 1
    for _ in range(steps_before_timeout):
        timeout(context)

    assert not context.is_active

    timeout.close()


def test_min_active():
    max_active = 500
    min_active = 120
    context = SpeechContext()
    timeout = ActivationTimeout(min_active=min_active, max_active=max_active)

    context.is_active = True

    # call with speech active
    context.is_speech = True
    timeout(context)

    # call timeout after speech is no longer detected
    context.is_speech = False
    timeout(context)
    assert context.is_active

    # vad fall should be True
    # with context still active
    timeout(context)
    assert context.is_active

    # context should remain active until min active
    steps_before_deactivate = min_active // 20
    for _ in range(steps_before_deactivate):
        timeout(context)
        assert context.is_active

    # call with speech active
    context.is_speech = True
    timeout(context)

    # call timeout after speech is no longer detected
    # min active should be satisfied
    context.is_speech = False
    timeout(context)
    assert not context.is_active

    timeout.close()



"""
This module contains the tests for the automatic noise suppression extension.
"""
import numpy as np
import pytest

from spokestack import utils
from spokestack.context import SpeechContext
from spokestack.nsx.webrtc import AutomaticNoiseSuppression

np.random.seed(42)


def test_construction():
    nsx = AutomaticNoiseSuppression(16000, 1)
    nsx.reset()
    nsx.close()


def test_invalid_sample_rate():
    with pytest.raises(ValueError):
        _ = AutomaticNoiseSuppression(sample_rate=900000, policy=1)


def test_invalid_frame_size():
    context = SpeechContext()
    nsx = AutomaticNoiseSuppression(16000, 1)

    bad_frame = np.random.rand(550)
    with pytest.raises(ValueError):
        nsx(context, bad_frame)


def test_invalid_frame_dtype():
    context = SpeechContext()
    nsx = AutomaticNoiseSuppression(16000, 1)

    bad_frame = np.random.rand(320)
    with pytest.raises(TypeError):
        nsx(context, bad_frame)


def test_processing():
    context = SpeechContext()
    nsx = AutomaticNoiseSuppression(16000, 1)

    # no suppression
    expect = sin_frame()
    actual = sin_frame()
    np.allclose(rms(expect), rms(actual), atol=3)

    nsx(context, utils.float_to_int16(sin_frame()))
    nsx(context, utils.float_to_int16(actual))
    np.allclose(rms(expect), rms(actual), atol=3)

    nsx.close()

    # valid suppression
    expect = sin_frame()
    actual = add_noise(sin_frame())
    nsx(context, utils.float_to_int16(sin_frame()))
    nsx(context, utils.float_to_int16(actual))
    np.allclose(rms(expect), rms(actual), atol=3)


def sin_frame(sample_rate=16000, frequency=100, frame_width=20):
    frame_width = sample_rate * frame_width // 1000
    x = 2 * np.pi * np.arange(sample_rate) / sample_rate
    frame = np.sin(frequency * x)
    frame = frame[:frame_width]
    return frame


def rms(y):
    return 20 * np.log10(max(np.sqrt(np.mean(y ** 2)), 1e-5) / 2e-5)


def add_noise(frame):
    noise = np.random.normal(size=frame.shape)
    noise = noise / np.power(10, 10 / 20.0)
    frame += noise
    return np.clip(frame, -1.0, 1.0)



"""
This module contains the websocket logic used to communicate with
Spokestack's cloud-based ASR service.
"""
import base64
import hashlib
import hmac
import json
from typing import Any, Dict, List, Union

import numpy as np
from websocket import WebSocket


class CloudClient:
    """Spokestack client for cloud based speech to text

    Args:
        key_id (str): identity from spokestack api credentials
        key_secret (str): secret key from spokestack api credentials
        socket_url (str): url for socket connection
        audio_format (str): format of input audio
        sample_rate (int): audio sample rate (kHz)
        language (str): language for recognition
        limit (int): Limit of messages per api response
        idle_timeout (Any): Time before client timeout. Defaults to None
    """

    def __init__(
        self,
        key_id: str,
        key_secret: str,
        socket_url: str = "wss://api.spokestack.io",
        audio_format: str = "PCM16LE",
        sample_rate: int = 16000,
        language: str = "en",
        limit: int = 10,
        idle_timeout: Union[float, None] = None,
    ) -> None:

        self._body: str = json.dumps(
            {
                "format": audio_format,
                "rate": sample_rate,
                "language": language,
                "limit": limit,
            }
        )
        self._socket_url: str = socket_url
        self._key_id: str = key_id
        self._key: bytes = key_secret.encode("utf-8")
        signature = hmac.new(
            self._key, self._body.encode("utf-8"), hashlib.sha256
        ).digest()
        self._signature = base64.b64encode(signature).decode("utf-8")
        self._socket: Any = None

        self._response: Dict[str, Any] = {
            "error": None,
            "final": True,
            "hypotheses": [],
            "status": None,
        }
        self._sample_rate: int = sample_rate
        self._idle_timeout = idle_timeout
        self._idle_count: int = 0

    def __call__(self, audio: Union[bytes, np.ndarray], limit: int = 1) -> List[str]:
        """Audio to text interface for the cloud client

        Args:
            audio (bytes|np.ndarray): input audio can be in the form of
                                      bytes or np.float, np.int16 array with
                                      conversions handled. other types with produce
                                      a TypeError
            limit (int): number of predictions to return

        Returns: list of transcripts, and their confidence values of size limit

        """
        if isinstance(audio, bytes):
            audio = np.frombuffer(audio, np.int16)
        elif np.issubdtype(audio.dtype, np.floating):
            # convert and rescale to PCM-16
            audio = (audio * (2 ** 15 - 1)).astype(np.int16)
        elif not np.issubdtype(audio.dtype, np.int16):
            raise TypeError("invalid_audio")

        chunk_size = self._sample_rate
        self.connect()
        self.initialize()

        for i in range(0, len(audio), chunk_size):
            frame = audio[i:][:chunk_size]
            self.send(frame)
            self.receive()

        self.end()
        while not self._response["final"]:
            self.receive()
        self.disconnect()

        hypotheses = self._response.get("hypotheses", [])
        return hypotheses[:limit]

    @property
    def is_connected(self) -> bool:
        """ status of the socket connection """
        if self._socket:
            return True
        return False

    def connect(self) -> None:
        """ connects to websocket """
        if self._socket is None:
            self._socket = WebSocket()
            self._socket.connect(f"{self._socket_url}/v1/asr/websocket")

    def initialize(self) -> None:
        """ sends/receives the initial api request """
        if not self._socket:
            raise ConnectionError("Not Connected")

        message = {
            "keyId": self._key_id,
            "signature": self._signature,
            "body": self._body,
        }
        self._socket.send(json.dumps(message))
        self._response = json.loads(self._socket.recv())
        if not self._response["status"] == "ok":
            raise APIError(self._response)

    def disconnect(self) -> None:
        """ disconnects client socket connection """
        if self._socket:
            self._socket.close()
            self._socket = None

    def send(self, frame: np.ndarray) -> None:
        """sends a single frame of audio

        Args:
            frame (np.ndarray): segment of PCM-16 encoded audio

        """
        if self._socket:
            self._socket.send_binary(frame.tobytes())
        else:
            raise ConnectionError("Not Connected")

    def end(self) -> None:
        """ sends empty string in binary to indicate last frame """
        if self._socket:
            self._socket.send_binary(b"")
        else:
            raise ConnectionError("Not Connected")

    def receive(self) -> None:
        """ receives the api response """
        if self._socket:
            timeout = self._socket.timeout
            try:
                self._socket.timeout = 0
                response = self._socket.recv()
                self._response = json.loads(response)
            except Exception:
                pass
            self._socket.timeout = timeout
        else:
            raise ConnectionError("Not Connected")

    @property
    def response(self) -> dict:
        """ current response message"""
        return self._response

    @property
    def is_final(self) -> bool:
        """ status of most recent sever response """
        return self._response["final"]

    @property
    def idle_timeout(self) -> Any:
        """ property for maximum idle time """
        return self._idle_timeout

    @property
    def idle_count(self) -> int:
        """ current counter of idle time """
        return self._idle_count

    @idle_count.setter
    def idle_count(self, value: int) -> None:
        """ sets the idle counter"""
        self._idle_count = value


class APIError(Exception):
    """Spokestack api error pass through

    Args:
        response (dict): message from the api service
    """

    def __init__(self, response: dict) -> None:
        super().__init__(response["error"])



"""
This module contains general utilities
"""
import json
from typing import Any, Dict

import numpy as np


def load_json(path: str) -> Dict[str, Any]:
    """

    Args:
        path (str): path to json file

    Returns: dictionary of json content

    """
    with open(path) as file:
        config = json.load(file)
    return config


def int16_to_float(audio: np.ndarray) -> np.ndarray:
    assert np.issubdtype(audio.dtype, np.int)
    return audio.astype(np.float32) / (2 ** 15 - 1)


def float_to_int16(audio: np.ndarray) -> np.ndarray:
    assert np.issubdtype(audio.dtype, np.floating)
    return (audio * (2 ** 15 - 1)).astype(np.int16)



"""
This module contains the class for using TFLite NLU models. In this case, an NLU model
is a TFLite model which takes in an utterance and returns an intent along with
any slots that are associated with that intent.
"""
import json
import logging
import os
from importlib import import_module
from typing import Any, Dict, List, Tuple

import numpy as np
from tokenizers import BertWordPieceTokenizer

from spokestack import utils
from spokestack.models.tensorflow import TFLiteModel
from spokestack.nlu.result import Result

_LOG = logging.getLogger(__name__)


class TFLiteNLU:
    """Abstraction for using TFLite NLU models

    Args:
        model_dir (str): path to the model directory containing nlu.tflite,
                         metadata.json, and vocab.txt
    """

    def __init__(self, model_dir: str) -> None:
        self._model = TFLiteModel(model_path=os.path.join(model_dir, "nlu.tflite"))
        self._metadata = utils.load_json(os.path.join(model_dir, "metadata.json"))
        self._tokenizer = BertWordPieceTokenizer(os.path.join(model_dir, "vocab.txt"))
        self._max_length = self._model.input_details[0]["shape"][-1]
        self._intent_decoder = {
            i: intent["name"] for i, intent in enumerate(self._metadata["intents"])
        }
        self._tag_decoder = {i: tag for i, tag in enumerate(self._metadata["tags"])}
        self._intent_meta = {
            intent.pop("name"): intent for intent in self._metadata["intents"]
        }
        self._slot_meta = {}
        for intent in self._intent_meta:
            for slot in self._intent_meta[intent]["slots"]:
                self._slot_meta[slot.pop("name")] = slot
        self._warm_up()

    def __call__(self, utterance: str) -> Result:
        """Classifies a string utterance into an intent and identifies any associated
            slots contained in the utterance. The slots get parsed based on type and
            then returned along with the intent and its associated confidence value.

        Args:
            utterance (str): string that needs to be understood

        Returns (Result): A class with properties for the identified intent, along with
                        raw, parsed slots and model confidence in prediction

        """
        inputs, input_ids = self._encode(utterance)
        outputs = self._model(inputs)
        intent, tags, confidence = self._decode(outputs)

        # slice off special tokens: [CLS], [SEP]
        tags = tags[: len(input_ids) - 2]
        _LOG.debug(f"{tags}")
        input_ids = input_ids[1:-1]
        _LOG.debug(f"{input_ids}")
        # retrieve slots from the tagged positions and decode slots back
        # into original values
        slots = [
            (token_id, tag[2:]) for token_id, tag in zip(input_ids, tags) if tag != "o"
        ]
        _LOG.debug(f"{slots}")

        slot_map: dict = {}
        for (token, tag) in slots:
            if tag in slot_map:
                slot_map[tag].append(token)
            else:
                slot_map[tag] = [token]

        for key, value in slot_map.items():
            slot_map[key] = self._tokenizer.decode(value)

        # attempt to resolve tagged tokens into slots and
        # collect the successful ones
        parsed_slots = {}
        for key in slot_map:
            parsed = self._parse_slots(self._slot_meta[key], slot_map[key])
            parsed_slots[key] = {
                "name": key,
                "parsed_value": parsed,
                "raw_value": slot_map[key],
            }
        _LOG.debug(f"parsed slots: {parsed_slots}")
        return Result(
            utterance=utterance,
            intent=intent,
            confidence=confidence,
            slots=parsed_slots,
        )

    def _warm_up(self) -> None:
        # make an array the same size as the inputs to warm the
        # model since first inference is always slower than subsequent
        warm = np.zeros((self._model.input_details[0]["shape"]), dtype=np.int32)
        _ = self._model(warm)

    def _encode(self, utterance: str) -> Tuple[np.ndarray, List[int]]:
        inputs = self._tokenizer.encode(utterance)
        # get the non-padded/truncated token ids to match the
        # original utterance to the respective labels and
        # use the length to slice the results
        input_ids = inputs.ids
        # it's (max_length + 1) because the [CLS]
        # token gets appended inside the model
        # notice the slice [1:] when we convert to an array
        inputs.truncate(max_length=self._max_length + 1)
        inputs.pad(length=self._max_length + 1)
        inputs = np.array(inputs.ids[1:], np.int32)
        # add the batch dimension for the TFLite model
        inputs = np.expand_dims(inputs, 0)
        return inputs, input_ids

    def _decode(self, outputs: list) -> Tuple[str, List[str], float]:
        # to get the index of the highest probability we
        # apply argmax to the posteriors which allows the
        # labels to be decoded with an integer to string mapping
        # we derive the confidence from the highest probability
        intent_posterior, tag_posterior = outputs
        intents, confidence = self._decode_intent(intent_posterior)
        tags = self._decode_tags(tag_posterior)
        _LOG.debug(f"decoded tags: {tags}")
        _LOG.debug(f"decoded intent: {intents}")
        _LOG.debug(f"confidence: {confidence}")
        return intents, tags, confidence

    def _decode_tags(self, posterior: np.ndarray) -> List[Any]:
        posterior = np.squeeze(posterior, 0)
        tags = np.argmax(posterior, -1)
        return [self._tag_decoder.get(tag) for tag in tags]

    def _decode_intent(self, posterior: np.ndarray) -> Any:
        posterior = np.squeeze(posterior, 0)
        intent = np.argmax(posterior, -1)
        return self._intent_decoder.get(intent), posterior[intent]

    def _parse_slots(self, slot_meta: Dict[str, Any], slots: Dict[str, Any]) -> Any:
        slot_type = slot_meta["type"]
        parser = import_module(f"spokestack.nlu.parsers.{slot_type}")
        facets = json.loads(slot_meta["facets"])
        return parser.parse(facets, slots)  # type: ignore



"""
This module contains the class for webrtc's automatic gain control
"""
from typing import Any

import numpy as np

from spokestack.context import SpeechContext
from spokestack.extensions.webrtc.agc import WebRtcAgc


class AutomaticGainControl:
    """WebRTC Automatic Gain Control

    Args:
            sample_rate (int): audio sample_rate. (Hz)
            frame_width (int): audio frame width. (Ms)
            target_level_dbfs (int): target peak audio level. (dBFS)
            compression_gain_db (int): dynamic range compression rate. (dB)
            limit_enable (bool): enables limiter in compression.
    """

    def __init__(
        self,
        sample_rate: int = 16000,
        frame_width: int = 20,
        target_level_dbfs: int = 3,
        compression_gain_db: int = 15,
        limit_enable: bool = True,
        **kwargs: Any
    ) -> None:
        # validate sample rate
        self._sample_rate = sample_rate
        if self._sample_rate not in {8000, 16000, 32000}:
            raise ValueError("invalid_sample_rate")
        self._frame_width = frame_width
        # validate frame width
        if self._frame_width not in {10, 20}:
            raise ValueError("invalid_frame_width")

        self._agc = WebRtcAgc(
            sample_rate=self._sample_rate,
            frame_width=self._frame_width,
            target_level_dbfs=target_level_dbfs,
            compression_gain_db=compression_gain_db,
            limit_enable=limit_enable,
        )

    def __call__(self, context: SpeechContext, frame: np.array) -> None:
        """Main Entry Point

        Args:
            context (SpeechContext): State based information that needs to be shared
            between pieces of the pipeline
            frame (np.array): PCM-16 audio.
        """
        # validate frame size
        if len(frame) != self._sample_rate * self._frame_width // 1000:
            raise ValueError("invalid_frame_size")
        # validate dtype
        if not np.issubdtype(frame.dtype, np.signedinteger):
            raise TypeError("invalid_dtype")
        # run automatic gain control on the frame
        self._agc(frame)

    def close(self) -> None:
        """method for pipeline compliance"""
        pass

    def reset(self) -> None:
        """method for pipeline compliance"""
        pass



"""
This module contains the tests for the wakeword asr profile.
"""
from unittest import mock

from spokestack.profile.wakeword_asr import WakewordSpokestackASR


@mock.patch("spokestack.profile.wakeword_asr.PyAudioInput")
@mock.patch("spokestack.profile.wakeword_asr.WakewordTrigger")
@mock.patch("spokestack.profile.wakeword_asr.SpeechPipeline")
def test_activate(*args):
    pipeline = WakewordSpokestackASR.create("", "")
    pipeline.start()
    pipeline.run()



"""
This module contains the tests for the TTSClient class
"""
from unittest import mock

import numpy as np
import pytest
from requests import Response

from spokestack.tts.clients.spokestack import TextToSpeechClient, TTSError


def test_synthesize_text():
    client = TextToSpeechClient("", "", "")

    test = np.ones(160).tobytes()
    with mock.patch("spokestack.tts.clients.spokestack.requests") as patched:
        mock_iterable = mock.MagicMock(
            spec=Response().iter_content(), return_value=test
        )
        patched.post.return_value = MockResponse(status_code=200)
        patched.get.return_value = mock.Mock(
            iter_content=mock_iterable, status_code=200
        )
        response = client.synthesize("test utterance")
        assert response == test


def test_synthesize_ssml():
    client = TextToSpeechClient("", "", "")

    test = np.ones(160).tobytes()
    with mock.patch("spokestack.tts.clients.spokestack.requests") as patched:
        mock_iterable = mock.MagicMock(
            spec=Response().iter_content(), return_value=test
        )
        patched.post.return_value = MockResponse(status_code=200)
        patched.get.return_value = mock.Mock(
            iter_content=mock_iterable, status_code=200
        )
        response = client.synthesize("<speak> test utterance </speak>", mode="ssml")
        assert response == test


def test_synthesize_markdown():
    client = TextToSpeechClient("", "", "")

    test = np.ones(160).tobytes()
    with mock.patch("spokestack.tts.clients.spokestack.requests") as patched:
        mock_iterable = mock.MagicMock(
            spec=Response().iter_content(), return_value=test
        )
        patched.post.return_value = MockResponse(status_code=200)
        patched.get.return_value = mock.Mock(
            iter_content=mock_iterable, status_code=200
        )
        response = client.synthesize("# test utterance", mode="markdown")
        assert response == test


def test_synthesize_url():
    client = TextToSpeechClient("", "", "")

    with mock.patch("spokestack.tts.clients.spokestack.requests") as patched:
        mock_url = "https://test"
        patched.post.return_value = MockResponse(
            status_code=200,
            return_value={"data": {"synthesizeText": {"url": mock_url}}},
        )
        response = client.synthesize_url("# test utterance", mode="text")
        assert response == mock_url


def test_synthesize_invalid_mode():
    client = TextToSpeechClient("", "", "")

    test = np.ones(160).tobytes()
    with mock.patch("spokestack.tts.clients.spokestack.requests") as patched:
        patched.get.return_value = mock.Mock(content=test)
        with pytest.raises(ValueError):
            _ = client.synthesize("test utterance", mode="python")


def test_error_response():
    client = TextToSpeechClient("", "", "")

    with mock.patch("spokestack.tts.clients.spokestack.requests") as patched:
        patched.post.return_value = MockResponse(
            status_code=200,
            return_value={
                "data": {"synthesizeSSML": None},
                "errors": [
                    {
                        "locations": [{"column": 0, "line": 3}],
                        "message": "synthesis_failed",
                        "path": ["synthesizeSSML"],
                    }
                ],
            },
        )

        with pytest.raises(TTSError):
            _ = client.synthesize("utterance")


def test_post_http_error():
    client = TextToSpeechClient("", "", "")

    with mock.patch("spokestack.tts.clients.spokestack.requests") as patched:
        patched.post.return_value = MockResponse(status_code=201)
        with pytest.raises(Exception):
            _ = client.synthesize("utterance")


def test_get_http_error():
    client = TextToSpeechClient("", "", "")

    with mock.patch("spokestack.tts.clients.spokestack.requests") as patched:
        patched.post.return_value = MockResponse(status_code=200)
        patched.get.return_value = MockResponse(status_code=201)
        with pytest.raises(Exception):
            _ = client.synthesize("utterance")


class MockResponse(mock.MagicMock):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def json(self):
        return self.return_value



"""
Tests for TFLite wakeword models
"""
from unittest import mock

import numpy as np
import pytest

from spokestack.context import SpeechContext
from spokestack.wakeword.tflite import WakewordTrigger


class ModelFactory(mock.MagicMock):
    def __call__(self, model_path):
        model = mock.MagicMock()
        if model_path.endswith("filter.tflite"):
            model.input_details = [{"shape": [1, 257]}]
            model.output_details = [{"shape": [1, 40]}]
            model.return_value = [np.zeros((1, 40))]
        elif model_path.endswith("encode.tflite"):
            model.input_details = [{"shape": [1, 1, 40]}, {"shape": [1, 128]}]
            model.output_details = [{"shape": [1, 128]}, {"shape": [1, 128]}]
            model.return_value = [np.zeros((1, 128)), np.zeros((1, 128))]
        elif model_path.endswith("detect.tflite"):
            model.input_details = [{"shape": [1, 100, 128]}]
            model.output_details = [{"shape": [1, 1]}]
            model.return_value = [np.zeros((1, 1))]
        return model


@mock.patch("spokestack.wakeword.tflite.TFLiteModel", new_callable=ModelFactory)
def test_invalid_args(_mock):
    with pytest.raises(ValueError):
        _ = WakewordTrigger(fft_window_type="hamming")


@mock.patch("spokestack.wakeword.tflite.TFLiteModel", new_callable=ModelFactory)
def test_detect_vad_inactive(_mock):
    context = SpeechContext()

    detector = WakewordTrigger(model_dir="wakeword_model")

    test_frame = np.random.rand(
        160,
    ).astype(np.float32)
    context.is_speech = False
    detector(context, test_frame)
    assert not context.is_active


@mock.patch("spokestack.wakeword.tflite.TFLiteModel", new_callable=ModelFactory)
def test_detect_vad_active(_mock):
    context = SpeechContext()

    detector = WakewordTrigger(model_dir="wakeword_model")

    for _ in range(1):
        test_frame = np.random.rand(
            160,
        ).astype(np.float32)
        context.is_speech = True
        detector(context, test_frame)
        assert not context.is_active

    detector.close()


@mock.patch("spokestack.wakeword.tflite.TFLiteModel", new_callable=ModelFactory)
def test_detect_inactive_vad_deactivate(_mock):
    context = SpeechContext()
    detector = WakewordTrigger(model_dir="wakeword_model")

    for _ in range(3):
        test_frame = np.random.rand(
            160,
        ).astype(np.float32)
        context.is_speech = True
        detector(context, test_frame)
        context.is_speech = False
        assert not context.is_active
    detector(context, test_frame)


@mock.patch("spokestack.wakeword.tflite.TFLiteModel", new_callable=ModelFactory)
def test_detect_activate(_mock):
    context = SpeechContext()
    detector = WakewordTrigger(model_dir="wakeword_model")
    detector.detect_model.return_value[0][:] = 0.6

    test_frame = np.random.rand(
        512,
    ).astype(np.float32)
    context.is_speech = True
    detector(context, test_frame)
    context.is_speech = False
    assert context.is_active


@mock.patch("spokestack.wakeword.tflite.TFLiteModel", new_callable=ModelFactory)
def test_detect_active_min_delay(_mock):
    context = SpeechContext()
    detector = WakewordTrigger(model_dir="wakeword_model")

    test_frame = np.random.rand(
        512,
    ).astype(np.float32)
    context.is_speech = True
    detector(context, test_frame)

    detector.detect_model.return_value[0][:] = 1
    detector(context, test_frame)

    context.is_speech = False
    detector(context, test_frame)
    detector(context, test_frame)
    assert context.is_active


@mock.patch("spokestack.wakeword.tflite.TFLiteModel", new_callable=ModelFactory)
def test_detect_manual_min_delay(_mock):
    context = SpeechContext()
    detector = WakewordTrigger(model_dir="wakeword_model")
    detector.detect_model.return_value[0][:] = 1

    context.is_active = True
    test_frame = np.random.rand(
        512,
    ).astype(np.float32)
    detector(context, test_frame)
    detector(context, test_frame)
    detector(context, test_frame)

    assert context.is_active



"""
This module contains the tests for the entity parser
"""
from spokestack.nlu.parsers import entity


def test_entity_real():
    metadata = {}
    raw_value = "im a real entity"
    parsed = entity.parse(metadata, raw_value)
    assert parsed == raw_value


def test_entity_empty():
    metadata = {}
    raw_value = ""
    parsed = entity.parse(metadata, raw_value)
    assert parsed == raw_value



"""
This module contains the Spokestack client for text to speech
"""
import base64
import hashlib
import hmac
import json
from typing import Any, Iterator

import requests

_MODES = {
    "ssml": "synthesizeSSML",
    "markdown": "synthesizeMarkdown",
    "text": "synthesizeText",
}


class TextToSpeechClient:
    """Spokestack Text to Speech Client

    Args:
        key_id (str): identity from spokestack api credentials
        key_secret (str): secret key from spokestack api credentials
        url (str): spokestack api url
    """

    def __init__(
        self, key_id: str, key_secret: str, url: str = "https://api.spokestack.io/v1"
    ) -> None:

        self._key_id = key_id
        self._key = key_secret.encode("utf-8")
        self._url = url

    def synthesize(
        self,
        utterance: str,
        mode: str = "text",
        voice: str = "demo-male",
    ) -> Iterator[bytes]:
        """Converts the given utterance to speech

        Args:
            utterance (str): string that needs to be rendered as speech.
            mode (str): synthesis mode to use with utterance. text, ssml, markdown.
            voice (str): name of the tts voice.

        Returns:
            (Iterator[bytes]): Encoded audio response in the form of a sequence of bytes

        """
        audio_url = self.synthesize_url(utterance, mode, voice)
        response = requests.get(audio_url, stream=True)

        if response.status_code != 200:
            raise Exception(response.reason)

        return response.iter_content(chunk_size=None)

    def synthesize_url(
        self,
        utterance: str,
        mode: str = "text",
        voice: str = "demo-male",
    ) -> str:
        """Converts the given uttrance to speech accessible by a URL.

        Args:
            utterance (str): string that needs to be rendered as speech.
            mode (str): synthesis mode to use with utterance. text, ssml, markdown.
            voice (str): name of the tts voice.

        Returns: URL of the audio clip

        """
        body = self._build_body(utterance, mode, voice)
        signature = base64.b64encode(
            hmac.new(self._key, body.encode("utf-8"), hashlib.sha256).digest()
        ).decode("utf-8")
        headers = {
            "Authorization": f"Spokestack {self._key_id}:{signature}",
            "Content-Type": "application/json",
        }
        response: Any = requests.post(self._url, headers=headers, data=body)

        if response.status_code != 200:
            raise Exception(response.reason)

        response = response.json()
        if "errors" in response:
            raise TTSError(response["errors"])

        return response["data"][_MODES[mode]]["url"]

    @staticmethod
    def _build_body(message: str, mode: str, voice: str) -> str:
        if mode == "ssml":
            return json.dumps(
                {
                    "query": """
               query synthesis($voice: String!, $ssml: String!) {
                 synthesizeSSML(voice: $voice, ssml: $ssml) {url}
               }
               """,
                    "variables": {"voice": voice, "ssml": message},
                }
            )
        elif mode == "markdown":
            return json.dumps(
                {
                    "query": """
                           query synthesis($voice: String!, $markdown: String!) {
                             synthesizeMarkdown(voice: $voice, markdown: $markdown) {
                             url}
                           }
                           """,
                    "variables": {
                        "voice": voice,
                        "markdown": message,
                        "method": "synthesizeMarkdown",
                    },
                }
            )
        elif mode == "text":
            return json.dumps(
                {
                    "query": """
                           query synthesis($voice: String!, $text: String!) {
                             synthesizeText(voice: $voice, text: $text) {url}
                           }
                           """,
                    "variables": {"voice": voice, "text": message},
                }
            )
        else:
            raise ValueError("invalid_mode")


class TTSError(Exception):
    """ Text to speech error wrapper """

    def __init__(self, response: Any) -> None:
        messages = [error["message"] for error in response]
        super().__init__(messages)



"""
This module contains the logic to parse selsets from NLU results. Selsets contain a name
along with one or more aliases. This allows one to map any of the listed aliases
into a single word. For example, if a selset's name is "light", and its aliases are
bulbs, light, beam, lamp, etc., occurrences of any alias will be parsed as light
"""
from typing import Any, Dict, Union


def parse(metadata: Dict[str, Any], raw_value: str) -> Union[str, None]:
    """Selset Parser

    Args:
        metadata (Dict[str, Any]): slot metadata
        raw_value (str): value tagged by the model

    Returns:
        Union[str, None]: selset or None if invalid
    """
    normalized = raw_value.lower()
    selections = metadata.get("selections", [])
    for selection in selections:
        name = selection.get("name")
        if name.lower() == normalized:
            return name
        aliases = selection.get("aliases")
        for alias in aliases:
            if alias.lower() == normalized:
                return name
    return None



"""
This module contains the tests for pyaudio input/output
"""
from unittest import mock

import numpy as np

from spokestack.io import pyaudio


@mock.patch("spokestack.io.pyaudio.pyaudio")
def test_input(*args):
    mic = pyaudio.PyAudioInput(sample_rate=16000, frame_width=10)
    mic._stream.read.return_value = np.zeros(160, np.int16).tobytes()
    mic._audio.open.assert_called()
    # start audio stream
    mic.start()
    assert mic.is_active
    # read a single frame
    _ = mic.read()
    mic._stream.read.assert_called()
    # stop audio stream
    mic.stop()
    assert mic.is_stopped
    mic.close()


@mock.patch("spokestack.io.pyaudio.pyaudio")
def test_output(*args):
    speaker = pyaudio.PyAudioOutput()
    audio = np.ones(160, np.int16).tobytes()
    speaker.write(audio)



"""
Tests for RingBuffer class
"""
import numpy as np
import pytest

from spokestack.ring_buffer import RingBuffer


def test_ring_buffer():
    buffer = RingBuffer([0, 1])
    assert buffer.capacity == 0
    assert buffer.is_empty
    assert buffer.is_full

    buffer = RingBuffer([1, 1])
    assert buffer.capacity == 1
    assert buffer.is_empty
    assert not buffer.is_full

    buffer = RingBuffer([10, 1])
    assert buffer.capacity == 10
    assert buffer.is_empty
    assert not buffer.is_full


def test_read_write():
    buffer = RingBuffer([3, 1])
    for i in range(buffer.capacity):
        buffer.write(np.ones((1, 1)) * (i + 1))

    assert not buffer.is_empty
    assert buffer.is_full

    # full write
    with pytest.raises(IndexError):
        buffer.write(np.ones((1, 1)))

    for i in range(buffer.capacity):
        assert buffer.read() == np.ones((1, 1)) * (i + 1)

    assert buffer.is_empty
    assert not buffer.is_full

    # empty read
    with pytest.raises(IndexError):
        buffer.read()


def test_rewind():
    buffer = RingBuffer([4, 1])

    # default
    buffer.rewind()
    assert not buffer.is_empty
    assert buffer.is_full
    while not buffer.is_empty:
        buffer.read()

    # valid
    for i in range(buffer.capacity):
        buffer.write(np.ones((1, 1)) * (i + 1))
    while not buffer.is_empty:
        buffer.read()

    buffer.rewind()
    assert not buffer.is_empty
    assert buffer.is_full

    for i in range(buffer.capacity):
        assert buffer.read() == np.ones((1, 1)) * (i + 1)

    assert buffer.is_empty
    assert not buffer.is_full


def test_seek():
    buffer = RingBuffer([5, 1])

    # valid
    for i in range(buffer.capacity):
        buffer.write(np.ones((1, 1)) * (i + 1))

    buffer.seek(1)
    for i in range(1, buffer.capacity):
        assert buffer.read() == np.ones((1, 1)) * (i + 1)
    buffer.rewind()

    buffer.seek(buffer.capacity - 1)
    for i in range((buffer.capacity - 1), buffer.capacity):
        assert buffer.read() == np.ones((1, 1)) * (i + 1)

    buffer.seek(-7)
    assert buffer.read() == np.ones((1, 1)) * buffer.capacity


def test_reset():
    buffer = RingBuffer([4, 1])
    buffer.fill(1)
    assert not buffer.is_empty
    assert buffer.is_full

    for i in range(buffer.capacity):
        assert buffer.read() == np.ones((1, 1), np.float32)

    buffer.reset()
    buffer.fill(1)
    assert not buffer.is_empty
    assert buffer.is_full

    for i in range(buffer.capacity - 1):
        assert buffer.read() == np.ones((1, 1))

    assert buffer.read() == np.ones((1, 1))


def test_read_all():
    buffer = RingBuffer([3, 1])
    values = []
    for i in range(buffer.capacity):
        inputs = np.ones((1, 1)) * (i + 1)
        buffer.write(inputs)
        values.append(inputs)

    assert not buffer.is_empty
    assert buffer.is_full

    all = buffer.read_all()
    assert buffer.is_empty
    assert not buffer.is_full
    assert all.all() == np.array(values).all()



"""
This module contains tests for SpeechContext
"""

from spokestack.context import SpeechContext


def test_context():
    context = SpeechContext()

    # test is_speech
    assert not context.is_speech
    context.is_speech = True
    assert context.is_speech

    # test is_active
    assert not context.is_active
    context.is_active = True
    assert context.is_active

    # test transcript
    assert not context.transcript
    context.transcript = "this is a test"
    assert context.transcript

    # test confidence
    assert context.confidence == 0.0
    context.confidence = 1.0
    assert context.confidence == 1.0

    # test reset
    context.reset()
    assert not context.is_speech
    assert not context.is_active
    assert not context.transcript
    assert context.confidence == 0.0


def test_handler():
    def on_speech(context):
        context.transcript = "event handled"

    context = SpeechContext()
    context.add_handler("recognize", on_speech)

    context.event("recognize")

    assert context.transcript == "event handled"



"""
This module contains the AutomaticGainControl class implemented in Cython
"""
# distutils: sources = filter_audio/agc/analog_agc.c, filter_audio/agc/digital_agc.c
# distutils: include_dirs = filter_audio/agc/include/
cimport numpy as np
cimport cagc

from spokestack.extensions.webrtc import ProcessError

MIC_MAX = 255
MIC_TARGET = 180


cdef class WebRtcAgc:
    cdef void* _agc

    def __dealloc__(self):
        cagc.WebRtcAgc_Free(self._agc)

    def __init__(self,
                 sample_rate,
                 frame_width,
                 target_level_dbfs,
                 compression_gain_db,
                 limit_enable
    ):
        self._agc = NULL
        result = cagc.WebRtcAgc_Create(&self._agc)
        if result != 0:
            raise MemoryError("out_of_memory")

        result = cagc.WebRtcAgc_Init(self._agc,
                                     minLevel=0,
                                     maxLevel=MIC_MAX,
                                     agcMode=2,
                                     fs=sample_rate
        )
        if result != 0:
            raise ValueError("invalid_config")

        cdef cagc.WebRtcAgc_config_t config
        config.targetLevelDbfs = target_level_dbfs
        config.limiterEnable = limit_enable
        config.compressionGaindB = compression_gain_db

        result = cagc.WebRtcAgc_set_config(self._agc, config)
        if result != 0:
            raise ValueError("invalid_config")

    def __call__(self, frame):
        self._process(frame)

    cdef _process(self, frame):
        cdef char saturated = 0
        cdef int mic_level = 0
        result = cagc.WebRtcAgc_VirtualMic(
            agcInst=self._agc,
            inMic=<short*> np.PyArray_DATA(frame),
            inMic_H=NULL,
            samples=<short> len(frame),
            micLevelIn=MIC_TARGET,
            micLevelOut=&mic_level
        )
        if result != 0:
            raise ProcessError("mic_failed")

        result = cagc.WebRtcAgc_Process(
                agcInst=self._agc,
                inNear=<short*> np.PyArray_DATA(frame),
                inNearH=NULL,
                samples=<short> len(frame),
                out=<short*> np.PyArray_DATA(frame),
                out_H=NULL,
                inMicLevel=MIC_TARGET,
                outMicLevel=&mic_level,
                echo=0,
                saturationWarning=&saturated
        )
        if result != 0:
            raise ProcessError("invalid_input")



"""
This module contains tests for the TFLiteNLU class
"""
import json
from unittest import mock

import numpy as np

from spokestack.nlu.tflite import TFLiteNLU


@mock.patch("spokestack.nlu.tflite.BertWordPieceTokenizer")
@mock.patch("spokestack.nlu.tflite.TFLiteModel")
def test_classify_without_slots(_mock_tokenizer, _mock_model, fs):
    metadata = json.dumps(
        {
            "domain": "dummy",
            "intents": [
                {
                    "name": "command.test",
                    "description": "",
                    "implicit_slots": [],
                    "slots": [],
                }
            ],
            "tags": ["o", "b_test", "i_test"],
        }
    )
    fs.create_file("/tmp/model/vocab.txt")
    fs.create_file("/tmp/model/metadata.json", contents=metadata)
    model = TFLiteNLU("/tmp/model/")

    model._model._input_details = [{"shape": [1, 50]}]
    model._model._output_details = [{"shape": [1, 1]}, {"shape": [1, 50, 3]}]
    model._model.return_value = [np.random.random((1, 1)), np.random.random((1, 50, 3))]

    utterance = "this is only a test"
    outputs = model(utterance)
    assert outputs.utterance == utterance
    assert 0.0 <= outputs.confidence <= 1.0
    assert not outputs.slots


@mock.patch("spokestack.nlu.tflite.BertWordPieceTokenizer")
@mock.patch("spokestack.nlu.tflite.TFLiteModel")
def test_classify_with_slots(_mock_tokenizer, _mock_model, fs):
    metadata = json.dumps(
        {
            "domain": "dummy",
            "intents": [
                {
                    "name": "command.test",
                    "description": "",
                    "implicit_slots": [],
                    "slots": [
                        {
                            "name": "test",
                            "capture_name": "test",
                            "description": "",
                            "type": "entity",
                            "facets": "{}",
                        }
                    ],
                }
            ],
            "tags": ["o", "b_test", "i_test"],
        }
    )
    fs.create_file("/tmp/model/vocab.txt")
    fs.create_file("/tmp/model/metadata.json", contents=metadata)
    model = TFLiteNLU("/tmp/model/")

    model._model._input_details = [{"shape": [1, 50]}]
    model._model._output_details = [{"shape": [1, 1]}, {"shape": [1, 50, 3]}]
    model._model.return_value = [np.random.random((1, 1)), np.random.random((1, 50, 3))]

    utterance = "this is only a test"
    model._encode = mock.MagicMock(return_value=[[utterance], [0, 1, 2, 3, 4, 0]])
    outputs = model(utterance)
    assert outputs.utterance == utterance
    assert 0.0 <= outputs.confidence <= 1.0
    assert outputs.slots


@mock.patch("spokestack.nlu.tflite.BertWordPieceTokenizer")
@mock.patch("spokestack.nlu.tflite.TFLiteModel")
def test_classify_with_multiple_slots(_mock_tokenizer, _mock_model, fs):
    metadata = json.dumps(
        {
            "domain": "dummy",
            "intents": [
                {
                    "name": "command.test",
                    "description": "",
                    "implicit_slots": [],
                    "slots": [
                        {
                            "name": "test",
                            "capture_name": "test",
                            "description": "",
                            "type": "entity",
                            "facets": "{}",
                        },
                        {
                            "name": "number",
                            "capture_name": "number",
                            "description": "",
                            "type": "entity",
                            "facets": "{}",
                        },
                    ],
                }
            ],
            "tags": ["o", "b_test", "i_test", "b_number", "i_number"],
        }
    )
    fs.create_file("/tmp/model/vocab.txt")
    fs.create_file("/tmp/model/metadata.json", contents=metadata)
    model = TFLiteNLU("/tmp/model/")

    model._model._input_details = [{"shape": [1, 5]}]
    model._model._output_details = [{"shape": [1, 1]}, {"shape": [1, 5, 5]}]
    model._model.return_value = [
        np.array([[1.0]]),
        np.array(
            [
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0],
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                ]
            ]
        ),
    ]
    model._tokenizer.decode.side_effect = ["test", "ninety nine"]
    utterance = "[CLS] this is test number ninety nine [SEP]"
    model._encode = mock.MagicMock(return_value=[[utterance], [0, 0, 0, 1, 0, 3, 4, 0]])
    outputs = model(utterance)

    assert outputs.utterance == utterance
    assert outputs.intent == "command.test"
    assert 0.0 <= outputs.confidence <= 1.0
    assert outputs.slots == {
        "test": {"name": "test", "parsed_value": "test", "raw_value": "test"},
        "number": {
            "name": "number",
            "parsed_value": "ninety nine",
            "raw_value": "ninety nine",
        },
    }



"""
This module contains the Automatic Gain Control's C API redefined in Cython.
"""



cdef extern from "filter_audio/agc/include/gain_control.h":
    ctypedef struct WebRtcAgc_config_t:
        int targetLevelDbfs
        int compressionGaindB
        int limiterEnable

    int WebRtcAgc_Create(void** agcInst)
    int WebRtcAgc_Init(void*agcInst,
                       int minLevel,
                       int maxLevel,
                       short agcMode,
                       int fs)

    int WebRtcAgc_set_config(void*agcInst, WebRtcAgc_config_t config)

    int WebRtcAgc_Free(void*agcInst)

    int WebRtcAgc_VirtualMic(void*agcInst,
                             short*inMic,
                             short*inMic_H,
                             short samples,
                             int micLevelIn,
                             int*micLevelOut)

    int WebRtcAgc_Process(void*agcInst,
                          short*inNear,
                          short*inNearH,
                          short samples,
                          short*out,
                          short*out_H,
                          int inMicLevel,
                          int*outMicLevel,
                          short echo,
                          char*saturationWarning)



"""
This module uses pyaudio for input and output processing
"""
from typing import Any

import numpy as np
import pyaudio


class PyAudioInput:
    """This class retrieves audio from an input device

    Args:
        sample_rate (int): desired sample rate for input (Hz)
        frame_width (int): desired frame width for input (ms)
        exception_on_overflow (bool): produce exception for input overflow
    """

    def __init__(
        self,
        sample_rate: int,
        frame_width: int,
        exception_on_overflow: bool = True,
        **kwargs: Any
    ) -> None:
        self._frame_size = int(sample_rate / 1000 * frame_width)
        self._exception_on_overflow = exception_on_overflow
        self._audio = pyaudio.PyAudio()
        device = self._audio.get_default_input_device_info()
        self._stream = self._audio.open(
            input=True,
            input_device_index=device["index"],
            format=pyaudio.paInt16,
            frames_per_buffer=self._frame_size,
            channels=1,
            rate=sample_rate,
            start=False,
        )

    def read(self) -> np.array:
        """Reads a single frame of audio

        Returns:
            np.ndarray: single frame of PCM-16 audio
        """
        frame = self._stream.read(
            self._frame_size, exception_on_overflow=self._exception_on_overflow
        )
        return np.frombuffer(frame, np.int16)

    def start(self) -> None:
        """ Starts the audio stream """
        self._stream.start_stream()

    def stop(self) -> None:
        """ Stops the audio stream """
        self._stream.stop_stream()

    def close(self) -> None:
        """ Closes the audio stream """
        self._stream.close()

    @property
    def is_active(self) -> bool:
        """Stream active property

        Returns:
            bool: 'True' if stream is active, 'False' otherwise
        """
        return self._stream.is_active()

    @property
    def is_stopped(self) -> bool:
        """Stream stopped property

        Returns:
            bool: 'True' if stream is stopped, 'False' otherwise
        """
        return self._stream.is_stopped()


class PyAudioOutput:
    """Outputs audio to the default system output

    Args:
        num_channels (int): number of audio channels
        sample_rate (int): sample rate of the audio (Hz)
    """

    def __init__(self, num_channels: int = 1, sample_rate: int = 24000) -> None:
        audio = pyaudio.PyAudio()
        device = audio.get_default_output_device_info()
        self._output = audio.open(
            output=True,
            input_device_index=device["index"],
            format=pyaudio.paInt16,
            channels=num_channels,
            rate=sample_rate,
        )

    def write(self, frame: bytes) -> None:
        """Writes a single frame of audio to output

        Args:
            frame (bytes): a single frame of audio

        """
        self._output.write(frame)



"""
This module contains the AutomaticNoiseSuppression class implemented in Cython
"""
# distutils: sources = filter_audio/ns/noise_suppression_x.c
# distutils: include_dirs = filter_audio/ns/include/
cimport cnsx
cimport numpy as np

from spokestack.extensions.webrtc import ProcessError


cdef class WebRtcNsx:
    cdef cnsx.NsxHandle* _ans

    def __dealloc__(self):
        cnsx.WebRtcNsx_Free(self._ans)
        self._ans = NULL

    def __init__(self, sample_rate, policy):
        self._ans = NULL

        result = cnsx.WebRtcNsx_Create(&self._ans)
        if result != 0:
            raise MemoryError("out_of_memory")

        result = cnsx.WebRtcNsx_Init(self._ans, sample_rate)
        if result != 0:
            raise ValueError("invalid_config")

        result = cnsx.WebRtcNsx_set_policy(self._ans, policy)
        if result != 0:
            raise ValueError("invalid_config")

    def __call__(self, frame):
        self._process(frame)

    cdef _process(self, frame):
        result = cnsx.WebRtcNsx_Process(self._ans,
                                        <short*> np.PyArray_DATA(frame),
                                        NULL,
                                        <short*> np.PyArray_DATA(frame),
                                        NULL)
        if result != 0:
            raise ProcessError("invalid_input")



"""
This module contains tests for Voice Activity Detection
"""
import numpy as np
import pytest

from spokestack import utils
from spokestack.context import SpeechContext
from spokestack.vad.webrtc import VoiceActivityDetector, VoiceActivityTrigger


def test_invalid_frame_width():
    with pytest.raises(ValueError):
        _ = VoiceActivityDetector(frame_width=30)


def test_invalid_sample_rate():
    with pytest.raises(ValueError):
        _ = VoiceActivityDetector(sample_rate=9000)


def test_invalid_dtype():
    context = SpeechContext()
    detector = VoiceActivityDetector()

    bad_frame = np.random.rand(160)
    with pytest.raises(Exception):
        detector(context, bad_frame)


def test_vad_is_triggered():
    context = SpeechContext()
    detector = VoiceActivityDetector(frame_width=10)

    frame = silence_frame()
    detector(context, frame)
    assert not context.is_speech

    frame = voice_frame()
    detector(context, frame)
    assert context.is_speech

    detector.close()


def test_vad_rise_delay():
    context = SpeechContext()
    detector = VoiceActivityDetector(frame_width=10, vad_rise_delay=30)
    for i in range(3):
        frame = voice_frame()
        detector(context, frame)
        if i < 2:
            assert not context.is_speech
        else:
            assert context.is_speech
    detector.close()


def test_vad_fall_triggered():
    context = SpeechContext()
    detector = VoiceActivityDetector(frame_width=10, vad_fall_delay=20)

    frame = voice_frame()
    detector(context, frame)
    assert context.is_speech

    frame = silence_frame()
    detector(context, frame)
    assert context.is_speech

    frame = voice_frame()
    detector(context, frame)
    assert context.is_speech

    detector.close()


def test_vad_fall_untriggered():
    context = SpeechContext()
    detector = VoiceActivityDetector(frame_width=10, vad_fall_delay=20)

    voice = voice_frame()
    silence = silence_frame()

    detector(context, voice)
    assert context.is_speech

    for i in range(10):
        detector(context, silence)
        assert context.is_speech

    detector(context, silence)
    assert not context.is_speech
    detector.close()


def test_voice_activity_trigger():
    context = SpeechContext()
    trigger = VoiceActivityTrigger()

    frame = voice_frame()

    trigger(context, frame)
    assert not context.is_active

    context.is_speech = True
    trigger(context, frame)
    assert context.is_active

    trigger.close()


def voice_frame(sample_rate=16000, frequency=2000, frame_width=10):
    frame_width = sample_rate * frame_width // 1000
    x = 2 * np.pi * np.arange(sample_rate) / sample_rate
    frame = np.sin(frequency * x)
    frame = frame[:frame_width]
    return utils.float_to_int16(frame)


def silence_frame():
    return np.zeros(160, np.int16)



"""
This module contains the tests for vad_trigger_asr profile
"""
from unittest import mock

from spokestack.profile.vad_trigger_asr import VoiceActivityTriggerSpokestackASR


@mock.patch("spokestack.profile.vad_trigger_asr.PyAudioInput")
@mock.patch("spokestack.profile.vad_trigger_asr.SpeechPipeline")
def test_activate(*args):
    pipeline = VoiceActivityTriggerSpokestackASR.create("", "")
    pipeline.start()
    pipeline.run()



"""
This module contains the tests for the GoogleSpeechRecognizer class
"""
from unittest import mock

import numpy as np

from spokestack.asr.google.speech_recognizer import GoogleSpeechRecognizer
from spokestack.context import SpeechContext


@mock.patch("spokestack.asr.google.speech_recognizer.speech")
@mock.patch("spokestack.asr.google.speech_recognizer.service_account")
def test_recognize(*args):
    context = SpeechContext()
    audio = np.zeros(160).astype(np.int16)
    recognizer = GoogleSpeechRecognizer(language="en-US", credentials="")

    context.is_active = True
    for i in range(10):
        if i > 3:
            context.is_active = False
        recognizer(context, audio)

    recognizer.reset()
    recognizer.close()


@mock.patch("spokestack.asr.google.speech_recognizer.speech")
@mock.patch("spokestack.asr.google.speech_recognizer.service_account")
def test_receive(*args):
    context = SpeechContext()
    audio = np.zeros(160).astype(np.int16)
    recognizer = GoogleSpeechRecognizer(language="en-US", credentials="")
    recognizer._queue.put([audio, audio, audio])

    recognizer._client.streaming_recognize.return_value = [
        mock.Mock(
            results=[
                mock.Mock(alternatives=[mock.Mock(transcript="test", confidence=0.99)])
            ]
        )
    ]

    context.is_active = True
    for i in range(10):
        if i > 3:

            context.is_active = False
        recognizer(context, audio)

    recognizer._thread = mock.Mock()
    recognizer.reset()
    recognizer.close()


@mock.patch("spokestack.asr.google.speech_recognizer.speech")
@mock.patch("spokestack.asr.google.speech_recognizer.service_account")
def test_drain(*args):
    audio = np.zeros(160).astype(np.int16)
    recognizer = GoogleSpeechRecognizer(language="en-US", credentials="")
    recognizer._queue.put([audio, audio, audio])
    next(recognizer._drain())



"""
This module contains the Spokestack text to speech manager which handles a
text to speech client, decodes the returned audio, and writes the audio to
the specified output.
"""
from typing import Any

from streamp3 import MP3Decoder


class TextToSpeechManager:
    """Manages tts client and io target.

    Args:
        client: Text to speech client that returns encoded mp3 audio
        output: Audio io target
    """

    def __init__(self, client: Any, output: Any) -> None:
        self._client = client
        self._output = output

    def synthesize(self, utterance: str, mode: str, voice: str) -> None:
        """Synthesizes the given utterance with the voice and format provided.

        Args:
            utterance (str): string that needs to be rendered as speech.
            mode (str): synthesis mode to use with utterance. text, ssml, markdown, etc.
            voice (str): name of the tts voice.

        """
        stream = self._client.synthesize(utterance, mode, voice)
        stream = SequenceIO(stream)
        for frame in MP3Decoder(stream):
            self._output.write(frame)

    def close(self) -> None:
        """ Closes the client and output. """
        self._client = None
        self._output = None


class SequenceIO:
    """ Wrapper that allows for incrementally received audio to be decoded. """

    def __init__(self, sequence: Any) -> None:
        self._sequence = iter(sequence)

    def read(self, size: Any = None) -> bytes:
        try:
            return next(self._sequence)
        except StopIteration:
            return b""



"""
This module contains the webrtc component for voice activity detection (vad)
"""
import logging
from typing import Any

import numpy as np

from spokestack.context import SpeechContext
from spokestack.extensions.webrtc.vad import WebRtcVad

QUALITY = 0
LOW_BITRATE = 1
AGGRESSIVE = 2
VERY_AGGRESSIVE = 3

_LOG = logging.getLogger(__name__)


class VoiceActivityDetector:
    """This class detects the presence of voice in a frame of audio.

    Args:
        sample_rate (int): sample rate of the audio (Hz)
        frame_width (int): width of the audio frame: 10, 20, or 30 (ms)
        vad_rise_delay (int): rising edge delay (ms)
        vad_fall_delay (int): falling edge delay (ms)
        mode (int): named constant to set mode for vad

    """

    def __init__(
        self,
        sample_rate: int = 16000,
        frame_width: int = 20,
        vad_rise_delay: int = 0,
        vad_fall_delay: int = 0,
        mode: int = QUALITY,
        **kwargs: Any
    ) -> None:

        # validate sample rate
        self._sample_rate: int = sample_rate
        if self._sample_rate not in {8000, 16000, 32000}:
            raise ValueError("invalid_sample_rate")
        self._frame_width: int = frame_width
        # validate frame width
        if self._frame_width not in {10, 20}:
            raise ValueError("invalid_frame_width")

        self._rise_length: int = vad_rise_delay // frame_width
        self._fall_length: int = vad_fall_delay // frame_width

        self._vad = WebRtcVad(sample_rate=sample_rate, mode=mode)

        self._run_value: int = 0
        self._run_length: int = 0

    def __call__(self, context: SpeechContext, frame: np.ndarray) -> None:
        """Processes a single frame of audio to determine if voice is present

        Args:
            context (SpeechContext): State based information that needs to be shared
            between pieces of the pipeline
            frame (np.ndarray): Single frame of PCM-16 audio from an input source

        """
        # validate dtype
        if not np.issubdtype(frame.dtype, np.signedinteger):
            raise TypeError("invalid_dtype")

        result: bool = self._vad.is_speech(frame)

        raw = result > 0
        if raw == self._run_value:
            self._run_length += 1
        else:
            self._run_value = raw
            self._run_length = 1

        if self._run_value != context.is_speech:
            if self._run_value and self._run_length >= self._rise_length:
                context.is_speech = True
                _LOG.info("vad: true")
            if not self._run_value and self._run_length >= self._fall_length:
                context.is_speech = False
                _LOG.info("vad: false")

    def reset(self) -> None:
        """ Resets the current state """
        self._run_value = 0
        self._run_length = 0

    def close(self) -> None:
        """ Close interface for use in pipeline """
        self.reset()


class VoiceActivityTrigger:
    """ Voice Activity Detector trigger pipeline component """

    def __init__(self) -> None:
        self._is_speech = False

    def __call__(self, context: SpeechContext, frame: np.ndarray) -> None:
        """Activates speech context whenever speech is detected

        Args:
            context (SpeechContext): State based information that needs to be shared
            between pieces of the pipeline
            frame (np.ndarray): Single frame of PCM-16 audio from an input source

        """
        if context.is_speech != self._is_speech:
            if context.is_speech:
                context.is_active = True
            self._is_speech = context.is_speech

    def close(self) -> None:
        self.reset()

    def reset(self) -> None:
        self._is_speech = False



"""
This module contains the tests for the selset parser
"""
from spokestack.nlu.parsers import selset


def test_selset():
    metadata = {
        "selections": [{"name": "lights", "aliases": ["lights", "beams", "bulbs"]}]
    }
    raw_value = "beams"
    parsed = selset.parse(metadata, raw_value)
    assert parsed == "lights"

    raw_value = "lights"
    parsed = selset.parse(metadata, raw_value)
    assert parsed == "lights"

    raw_value = "cat"
    parsed = selset.parse(metadata, raw_value)
    assert not parsed



"""
This module contains tests for the keyword recognizer.
"""
from unittest import mock

import numpy as np
import pytest

from spokestack.asr.keyword.tflite import KeywordRecognizer
from spokestack.context import SpeechContext


class ModelFactory(mock.MagicMock):
    def __call__(self, model_path):
        model = mock.MagicMock()
        if model_path.endswith("filter.tflite"):
            model.input_details = [{"shape": [1, 257]}]
            model.output_details = [{"shape": [1, 40]}]
            model.return_value = [np.zeros((1, 40))]
        elif model_path.endswith("encode.tflite"):
            model.input_details = [{"shape": [1, 1, 40]}, {"shape": [1, 128]}]
            model.output_details = [{"shape": [1, 128]}, {"shape": [1, 128]}]
            model.return_value = [np.zeros((1, 128)), np.zeros((1, 128))]
        elif model_path.endswith("detect.tflite"):
            model.input_details = [{"shape": [1, 92, 128]}]
            model.output_details = [{"shape": [1, 3]}]
            model.return_value = [[[0.8, 0.1, 0.3]]]
        return model


@mock.patch("spokestack.asr.keyword.tflite.TFLiteModel", new_callable=ModelFactory)
def test_invalid_args(*args):
    with pytest.raises(ValueError):
        _ = KeywordRecognizer(
            classes=["one", "two", "three"], fft_window_type="hamming"
        )


@mock.patch("spokestack.asr.keyword.tflite.TFLiteModel", new_callable=ModelFactory)
def test_invalid_classes(*args):
    with pytest.raises(ValueError):
        _ = KeywordRecognizer(classes=["one", "two", "three", "four"])

    with pytest.raises(ValueError):
        _ = KeywordRecognizer(classes=["one", "two"])


@mock.patch("spokestack.asr.keyword.tflite.TFLiteModel", new_callable=ModelFactory)
def test_recognize(*args):
    context = SpeechContext()
    recognizer = KeywordRecognizer(classes=["one", "two", "three"])

    test_frame = np.random.rand(
        160,
    ).astype(np.float32)

    context.is_active = True
    for i in range(10):
        recognizer(context, test_frame)
        recognizer(context, test_frame)

    context.is_active = False
    recognizer(context, test_frame)
    assert context.transcript == "one"

    recognizer.close()


@mock.patch("spokestack.asr.keyword.tflite.TFLiteModel", new_callable=ModelFactory)
def test_timeout(*args):
    context = SpeechContext()
    recognizer = KeywordRecognizer(classes=["one", "two", "three"])
    recognizer.detect_model.return_value = [[[0.0, 0.0, 0.0]]]

    test_frame = np.random.rand(
        160,
    ).astype(np.float32)

    context.is_active = True
    for i in range(10):
        recognizer(context, test_frame)
        recognizer(context, test_frame)

    context.is_active = False
    recognizer(context, test_frame)
    assert not context.transcript

    recognizer.close()



"""
This module contains the tests for the AutomaticGainControl extension
"""

import numpy as np
import pytest

from spokestack.agc.webrtc import AutomaticGainControl
from spokestack.context import SpeechContext


def test_construction():
    agc = AutomaticGainControl(8000, 10)
    agc.reset()
    agc.close()


def test_invalid_sample_rate():
    with pytest.raises(ValueError):
        _ = AutomaticGainControl(55000)


def test_invalid_frame_width():
    with pytest.raises(ValueError):
        _ = AutomaticGainControl(frame_width=9001)


def test_invalid_frame_dtype():
    context = SpeechContext()
    agc = AutomaticGainControl()

    bad_frame = np.random.rand(320)
    with pytest.raises(TypeError):
        agc(context, bad_frame)


def test_invalid_frame_size():
    context = SpeechContext()
    agc = AutomaticGainControl()

    bad_frame = np.random.rand(100)
    with pytest.raises(ValueError):
        agc(context, bad_frame)


def test_processing():
    context = SpeechContext()

    sample_rate = 8000
    frequency = 2000

    agc = AutomaticGainControl(
        sample_rate=sample_rate,
        frame_width=10,
        target_level_dbfs=9,
        compression_gain_db=15,
    )

    # valid amplification
    frame = sin_frame(sample_rate, frequency, amplitude=0.08)
    level = rms(frame)
    agc(context, frame)
    assert rms(frame) > level

    # valid attenuation
    frame = sin_frame(sample_rate, frequency)
    level = rms(frame)
    agc(context, frame)
    assert rms(frame) < level

    agc.close()


def sin_frame(sample_rate=16000, frequency=2000, amplitude=1.0):
    frame_width = sample_rate * 10 // 1000
    x = 2 * np.pi * np.arange(sample_rate) / sample_rate
    frame = (np.sin(frequency * x) * amplitude * 32768.0).astype(np.int16)
    return frame[:frame_width]


def rms(y):
    return np.sqrt(np.mean(y.astype(np.float32) ** 2))



"""
This module contains mappings to convert numbers in words
to integers. ie. five -> 5
"""

from typing import Dict


ENG_ZERO: Dict[str, int] = {
    "zero": 0,
    "oh": 0,
    "owe": 0,
}
ENG_MOD10: Dict[str, int] = {
    "one": 1,
    "won": 1,
    "two": 2,
    "too": 2,
    "to": 2,
    "three": 3,
    "four": 4,
    "for": 4,
    "fore": 4,
    "five": 5,
    "six": 6,
    "sicks": 6,
    "sics": 6,
    "seven": 7,
    "eight": 8,
    "ate": 8,
    "nine": 9,
}
ENG_MOD20: Dict[str, int] = {
    "ten": 10,
    "tin": 10,
    "eleven": 11,
    "twelve": 12,
    "thirteen": 13,
    "fourteen": 14,
    "fifteen": 15,
    "sixteen": 16,
    "seventeen": 17,
    "eighteen": 18,
    "nineteen": 19,
}

ENG_DIV10: Dict[str, int] = {
    "twenty": 2,
    "thirty": 3,
    "forty": 4,
    "fifty": 5,
    "sixty": 6,
    "seventy": 7,
    "eighty": 8,
    "ninety": 9,
}

ENG_EXP10: Dict[str, int] = {"hundred": 2, "thousand": 3}

MULTIPLIERS: Dict[str, int] = {
    "hundred": 100,
    "thousand": 1000,
    "million": int(1e6),
    "billion": int(1e9),
}
WORD_TO_NUM: Dict[str, int] = {
    **ENG_ZERO,
    **ENG_MOD10,
    **ENG_MOD20,
    **ENG_DIV10,
    **MULTIPLIERS,
    "first": 1,
    "second": 2,
    "third": 3,
    "fif": 5,
    "eigh": 8,
    "nin": 9,
    "twelf": 12,
    "twentie": 20,
    "thirtie": 30,
    "fortie": 40,
    "fiftie": 50,
    "seventie": 70,
    "eightie": 80,
    "ninetie": 90,
}



from unittest import mock

from spokestack.pipeline import SpeechPipeline


def test_start_stop():
    stages = [
        mock.MagicMock(),
        mock.MagicMock(),
        mock.MagicMock(),
    ]
    pipeline = SpeechPipeline(mock.MagicMock(), stages=stages)

    pipeline.start()
    assert pipeline.is_running

    # test second call to start, which ignores if the pipeline is running
    pipeline.start()

    pipeline.step()

    pipeline.close()
    assert not pipeline.is_running


def test_dispatch():
    stages = [
        mock.MagicMock(),
        mock.MagicMock(),
        mock.MagicMock(),
    ]
    pipeline = SpeechPipeline(mock.MagicMock(), stages=stages)

    pipeline.start()

    pipeline._dispatch()

    pipeline.is_managed = True

    pipeline._dispatch()


def test_activate_deactivate():
    stages = [
        mock.MagicMock(),
        mock.MagicMock(),
        mock.MagicMock(),
    ]
    pipeline = SpeechPipeline(mock.MagicMock(), stages=stages)

    pipeline.start()
    pipeline.activate()
    assert pipeline.context.is_active

    pipeline.deactivate()
    assert not pipeline.context.is_active


def test_cleanup():
    stages = [
        mock.MagicMock(),
        mock.MagicMock(),
        mock.MagicMock(),
    ]
    pipeline = SpeechPipeline(mock.MagicMock(), stages=stages)

    pipeline.start()
    assert pipeline.is_running

    pipeline.stop()
    assert not pipeline.is_running

    # run after stopped will trigger clean up
    pipeline.run()
    assert not pipeline._stages
    assert not pipeline._input_source


def test_events():
    stages = [
        mock.MagicMock(),
        mock.MagicMock(),
        mock.MagicMock(),
    ]
    pipeline = SpeechPipeline(mock.MagicMock(), stages=stages)

    @pipeline.event
    def on_speech(context):
        context.transcript = "event triggered"

    # test empty event
    pipeline.event(name="dummy_event")

    pipeline.context.event("speech")
    assert pipeline.context.transcript == "event triggered"


def test_run():
    stages = [
        mock.MagicMock(),
        mock.MagicMock(),
        mock.MagicMock(),
    ]
    pipeline = SpeechPipeline(mock.MagicMock(), stages=stages)

    @pipeline.event
    def on_step(context):
        pipeline.stop()

    pipeline.start()
    pipeline.run()


def test_pause_resume():
    stages = [
        mock.MagicMock(),
        mock.MagicMock(),
        mock.MagicMock(),
    ]
    pipeline = SpeechPipeline(mock.MagicMock(), stages=stages)

    pipeline.start()
    assert pipeline.is_running

    pipeline.step()
    pipeline.pause()
    pipeline._input_source.stop.assert_called()

    # verify it does nothing
    pipeline.step()

    pipeline.resume()
    pipeline._input_source.start.assert_called()
    pipeline.close()



cdef extern from "filter_audio/ns/include/noise_suppression_x.h":
    ctypedef struct NsxHandle:
        pass

    int WebRtcNsx_Create(NsxHandle** nsxInst)
    int WebRtcNsx_Free(NsxHandle*nsxInst)
    int WebRtcNsx_Init(NsxHandle*nsxInst, int fs)
    int WebRtcNsx_set_policy(NsxHandle*nsxInst, int mode)
    int WebRtcNsx_Process(NsxHandle*nsxInst,
                          short*speechFrame,
                          short*speechFrameHB,
                          short*outFrame,
                          short*outFrameHB)



"""
This module contains the Result class for the NLU.
"""
from typing import Any, Dict


class Result:
    """Convenience wrapper for NLU Results

    Args:
        utterance (str): original input string
        intent (str): detected user intention
        confidence (float): model confidence in intent prediction
        slots (Dict[str, Any]): specific tokens needed by intent and
                                type information necessary for parsing

    """

    def __init__(
        self, utterance: str, intent: str, confidence: float, slots: Dict[str, Any]
    ) -> None:
        self._utterance: str = utterance
        self._intent: str = intent
        self._confidence: float = confidence
        self._slots: Dict[str, Any] = slots

    @property
    def utterance(self) -> str:
        """ Transcript passed to the NLU model """
        return self._utterance

    @property
    def intent(self) -> str:
        """ Classified user intent """
        return self._intent

    @property
    def confidence(self) -> float:
        """ Model confidence in intent classification """
        return self._confidence

    @property
    def slots(self) -> Dict[str, Any]:
        """ Slots found by model """
        return self._slots



"""
This module implements the RingBuffer class
"""
from typing import Any, Union

import numpy as np


class RingBuffer:
    """ ring buffer """

    def __init__(self, shape: list, dtype: Any = np.float32) -> None:
        self._dtype = dtype
        self._shape = shape
        self._shape[0] += 1
        self._buffer = np.empty(shape=self._shape, dtype=self._dtype)
        self._read: int = 0
        self._write: int = 0
        self._max_length = self._buffer.shape[0]

    @property
    def is_empty(self) -> bool:
        """Determines if buffer is empty

        Returns: True if empty, False otherwise

        """
        return self._read == self._write

    @property
    def is_full(self) -> bool:
        """Determines if the buffer is full

        Returns: True if full, False otherwise

        """
        return self._read == (self._write + 1) % self._max_length

    @property
    def capacity(self) -> int:
        """The capacity of the buffer

        Returns: Max size of the buffer

        """
        return self._max_length - 1

    def rewind(self) -> Any:
        """Rewinds the read head of the buffer to the most recent start position

        Returns: self

        """
        self._read = (self._write + 1) % self._max_length
        return self

    def reset(self) -> Any:
        """Empties the buffer

        Returns: self

        """
        self._write = self._read
        return self

    def fill(self, value: Union[int, float]) -> Any:
        """Fills the with a specific value

        Args:
            value (int or float): Fill value for the buffer

        Returns: self

        """
        self._buffer.fill(value)
        self._read = (self._write + 1) % self._max_length
        return self

    def seek(self, steps: int) -> Any:
        """Moves the read head a specified number of steps

        Args:
            steps (int): desired step length

        Returns: self

        """
        self._read = (self._read + steps) % self._max_length
        return self

    def write(self, item: np.ndarray) -> None:
        """Writes to the buffer and advances write head

        Args:
            item (np.ndarray): Array to be written to the buffer. Can be n-dimensional

        Returns: None

        """
        if self.is_full:
            raise IndexError("Buffer is full")

        self._buffer[self._write] = item
        self._write = (self._write + 1) % self._max_length

    def read(self) -> np.ndarray:
        """Reads from the buffer and advances read head.

        Returns: Array at the current read position

        """
        if self.is_empty:
            raise IndexError("Buffer is empty")

        item = self._buffer[self._read : self._read + 1]
        self._read = (self._read + 1) % self._max_length
        return item

    def read_all(self) -> np.ndarray:
        """Dumps the entire contents of the buffer to an array

        Returns: Array with full contents of the buffer

        """

        self.rewind()
        current = []
        while not self.is_empty:
            current.append(self.read())
        return np.concatenate(current).astype(self._dtype)



"""
This module contains the tests for the integer parser
"""
from spokestack.nlu.parsers import integer


def test_integer():
    metadata = {"range": [1, 50]}
    raw_value = "one"
    parsed = integer.parse(metadata, raw_value)
    assert parsed == 1

    raw_value = "1"
    parsed = integer.parse(metadata, raw_value)
    assert parsed == 1

    raw_value = "oneth"
    parsed = integer.parse(metadata, raw_value)
    assert parsed == 1

    metadata = {"range": [1, 500]}
    raw_value = "four hundred four"
    parsed = integer.parse(metadata, raw_value)
    assert parsed == 404


def test_exponent():
    metadata = {"range": [1, 10000]}
    raw_value = "nine thousand one"
    parsed = integer.parse(metadata, raw_value)
    assert parsed == 9001

    raw_value = "one hundred"
    parsed = integer.parse(metadata, raw_value)
    assert parsed == 100


def test_teens():
    metadata = {"range": [1, 19]}
    raw_value = "thirteen"
    parsed = integer.parse(metadata, raw_value)
    assert parsed == 13

    raw_value = "13"
    parsed = integer.parse(metadata, raw_value)
    assert parsed == 13

    raw_value = "fourteen"
    parsed = integer.parse(metadata, raw_value)
    assert parsed == 14

    raw_value = "fifteen"
    parsed = integer.parse(metadata, raw_value)
    assert parsed == 15

    raw_value = "sixteen"
    parsed = integer.parse(metadata, raw_value)
    assert parsed == 16


def test_multiples_of_ten():
    metadata = {"range": [1, 100]}
    raw_value = "ten"
    parsed = integer.parse(metadata, raw_value)
    assert parsed == 10

    raw_value = "10"
    parsed = integer.parse(metadata, raw_value)
    assert parsed == 10

    raw_value = "20"
    parsed = integer.parse(metadata, raw_value)
    assert parsed == 20

    raw_value = "30"
    parsed = integer.parse(metadata, raw_value)
    assert parsed == 30

    raw_value = "44"
    parsed = integer.parse(metadata, raw_value)
    assert parsed == 44


def test_out_of_range():
    metadata = {"range": [1, 11]}
    raw_value = "twelve"
    parsed = integer.parse(metadata, raw_value)
    assert not parsed

    raw_value = "12"
    parsed = integer.parse(metadata, raw_value)
    assert not parsed

    raw_value = "100000000000"
    parsed = integer.parse(metadata, raw_value)
    assert not parsed


def test_invalid():
    metadata = {"range": [1, 400]}
    raw_value = "four hundred hundred"
    parsed = integer.parse(metadata, raw_value)
    assert not parsed

    raw_value = "word"
    parsed = integer.parse(metadata, raw_value)
    assert not parsed



"""
This module contains the tests for the digits parser
"""
from spokestack.nlu.parsers import digits


def test_digits():
    metadata = {"count": 7}
    raw_value = "eight six seven five three oh nine"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "8675309"

    raw_value = "one eight hundred six seven eight"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "1800678"

    raw_value = "eight sixty seven five three oh nine"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "8675309"

    metadata = {"count": 4}
    raw_value = "twenty three nineteen"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "2319"

    metadata = {"count": 6}
    raw_value = "sixty five thousand one"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "650001"

    metadata = {"count": 2}
    raw_value = "sixty"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "60"

    raw_value = ""
    parsed = digits.parse(metadata, raw_value)
    assert not parsed


def test_mod_twenty():
    metadata = {"count": 1}
    raw_values_single_digits = [
        "one",
        "two",
        "three",
        "four",
        "five",
        "six",
        "seven",
        "eight",
        "nine",
    ]
    raw_values_double_digits = [
        "ten",
        "eleven",
        "twelve",
        "thirteen",
        "fourteen",
        "fifteen",
        "sixteen",
        "seventeen",
        "eighteen",
        "nineteen",
    ]

    for i, raw_value in enumerate(raw_values_single_digits):
        parsed = digits.parse(metadata, raw_value)
        assert parsed == str(i + 1)

    metadata = {"count": 2}
    for i, raw_value in enumerate(raw_values_double_digits):
        parsed = digits.parse(metadata, raw_value)
        assert parsed == str(i + 10)


def test_div_ten():
    metadata = {"count": 2}
    raw_value = "twenty"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "20"

    raw_value = "forty one"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "41"

    raw_value = "fifty-two"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "52"

    raw_value = "sixty - eight"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "68"

    metadata = {"count": 4}
    raw_value = "thirty zero one"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "3001"

    raw_value = "seventy ten"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "7010"


def test_hundreds():
    metadata = {"count": 3}
    raw_value = "one hundred"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "100"

    metadata = {"count": 11}
    raw_value = "one eight hundred three three five twenty-two eleven"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "18003352211"

    raw_value = "18003352211"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "18003352211"


def test_thousands():
    metadata = {"count": 4}
    raw_value = "two thousand"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "2000"

    metadata = {"count": 11}
    raw_value = "one eight hundred three thirty-five four thousand"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "18003354000"


def test_homophones():
    metadata = {"count": 4}
    raw_value = "zero oh oh owe"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "0000"

    raw_value = "ten tin"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "1010"

    metadata = {"count": 2}
    raw_value = "one won"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "11"

    raw_value = "eight ate"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "88"

    metadata = {"count": 3}
    raw_value = "two too to"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "222"

    raw_value = "four for fore"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "444"

    raw_value = "six sicks sics"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "666"

    metadata = {"count": 12}
    raw_value = "oh one 23 for five 66 seventy-seven ate 9"
    parsed = digits.parse(metadata, raw_value)
    assert parsed == "012345667789"


def test_metadata_violation():
    metadata = {"count": 8}
    raw_value = "zero oh oh owe two two"
    parsed = digits.parse(metadata, raw_value)
    assert not parsed

    metadata = {"count": 1}
    parsed = digits.parse(metadata, raw_value)
    assert not parsed


def test_empty():
    metadata = {"count": 1}
    raw_value = "words"
    parsed = digits.parse(metadata, raw_value)
    assert not parsed



"""
This module contains the recognizer for cloud based ASR in
the speech pipeline
"""
import logging
from typing import Any

import numpy as np

from spokestack.asr.spokestack.cloud_client import CloudClient
from spokestack.context import SpeechContext

_LOG = logging.getLogger(__name__)


class CloudSpeechRecognizer:
    """Speech recognizer for use in the speech pipeline

    Args:
        spokestack_id (str): identity under spokestack api credentials
        spokestack_secret (str): secret key from spokestack api credentials
        language (str): language recognized
        sample_rate (int): audio sample rate (kHz)
        frame_width (int): frame width of the audio (ms)
        idle_timeout (int): the number of iterations before the connection times out
    """

    def __init__(
        self,
        spokestack_id: str = "",
        spokestack_secret: str = "",
        language: str = "en",
        sample_rate: int = 16000,
        frame_width: int = 20,
        idle_timeout: int = 5000,
        **kwargs: Any,
    ) -> None:

        self._client: CloudClient = CloudClient(
            key_id=spokestack_id,
            key_secret=spokestack_secret,
            language=language,
            sample_rate=sample_rate,
            idle_timeout=int(idle_timeout / frame_width),
        )
        self._is_active = False

    def __call__(self, context: SpeechContext, frame: np.ndarray) -> None:
        """Entry point of the recognizer

        Args:
            context (SpeechContext): current state of the speech pipeline
            frame (np.ndarray): single frame of audio

        """

        if context.is_active and not self._is_active:
            self._begin()
            self._send(frame)
            _LOG.debug("ready for speech")
        elif context.is_active:
            self._send(frame)
            self._receive(context)
            _LOG.debug("begin speech")
        elif self._is_active:
            self._commit()
            _LOG.debug("end speech")
        elif not self._client.is_final:
            self._receive(context)
        elif self._client.idle_count < self._client.idle_timeout:
            self._client.idle_count += 1
        else:
            self._client.disconnect()

    def _begin(self) -> None:
        self._client.connect()
        self._client.initialize()
        self._is_active = True
        self._client.idle_count = 0

    def _send(self, frame: np.ndarray) -> None:
        self._client.send(frame)

    def _receive(self, context: SpeechContext) -> None:
        self._client.receive()
        hypotheses = self._client.response.get("hypotheses")
        if hypotheses:
            hypothesis = hypotheses[0]
            context.transcript = hypothesis["transcript"]
            context.confidence = hypothesis["confidence"]
            if context.transcript:
                context.event("partial_recognize")

        if self._client.is_final:
            if context.transcript:
                context.event("recognize")
                _LOG.debug("recognize event")
            else:
                context.event("timeout")
                _LOG.debug("timeout event")

    def _commit(self) -> None:
        self._is_active = False
        self._client.end()

    def reset(self) -> None:
        """ resets client connection """
        self._client.idle_count = 0
        self._is_active = False
        self.close()

    def close(self) -> None:
        """ closes client connection """
        self._client.disconnect()



cimport cvad
cimport numpy as np

from spokestack.extensions.webrtc import ProcessError


cdef class WebRtcVad:
    cdef cvad.VadInst* _vad
    cdef int _sample_rate

    def __dealloc__(self):
        cvad.WebRtcVad_Free(self._vad)

    def __init__(self, sample_rate, mode):
        self._vad = NULL
        self._sample_rate = sample_rate

        result = cvad.WebRtcVad_Create(&self._vad)
        if result != 0:
            raise MemoryError("out_of_memory")

        result = cvad.WebRtcVad_Init(self._vad)
        if result != 0:
            raise ValueError("invalid_config")

        result = cvad.WebRtcVad_set_mode(self._vad, mode)
        if result != 0:
            raise ValueError("invalid_config")

    def is_speech(self, frame):
        result = cvad.WebRtcVad_Process(
            self._vad,
            self._sample_rate,
            <short*> np.PyArray_DATA(frame),
            len(frame)
        )
        return result



import os

import numpy as np
import setuptools
from Cython.Build import cythonize

SOURCES = [
    os.path.join("spokestack/extensions/webrtc", source)
    for source in [
        "filter_audio/other/complex_bit_reverse.c",
        "filter_audio/other/complex_fft.c",
        "filter_audio/other/copy_set_operations.c",
        "filter_audio/other/cross_correlation.c",
        "filter_audio/other/division_operations.c",
        "filter_audio/other/dot_product_with_scale.c",
        "filter_audio/other/downsample_fast.c",
        "filter_audio/other/energy.c",
        "filter_audio/other/get_scaling_square.c",
        "filter_audio/other/min_max_operations.c",
        "filter_audio/other/real_fft.c",
        "filter_audio/other/resample_by_2.c",
        "filter_audio/other/resample_by_2_internal.c",
        "filter_audio/other/resample_fractional.c",
        "filter_audio/other/resample_48khz.c",
        "filter_audio/other/spl_init.c",
        "filter_audio/other/spl_sqrt.c",
        "filter_audio/other/spl_sqrt_floor.c",
        "filter_audio/other/vector_scaling_operations.c",
        "filter_audio/vad/vad_core.c",
        "filter_audio/vad/vad_filterbank.c",
        "filter_audio/vad/vad_gmm.c",
        "filter_audio/vad/vad_sp.c",
        "filter_audio/vad/webrtc_vad.c",
        "filter_audio/agc/analog_agc.c",
        "filter_audio/agc/digital_agc.c",
        "filter_audio/ns/nsx_core.c",
        "filter_audio/ns/nsx_core_c.c",
        "filter_audio/ns/noise_suppression_x.c",
    ]
]

EXTENSIONS = [
    setuptools.Extension(
        "spokestack.extensions.webrtc.agc",
        ["spokestack/extensions/webrtc/agc.pyx"] + SOURCES,
        include_dirs=["filter_audio/agc/include/"],
    ),
    setuptools.Extension(
        "spokestack.extensions.webrtc.nsx",
        ["spokestack/extensions/webrtc/nsx.pyx"] + SOURCES,
        include_dirs=["filter_audio/ns/include/"],
    ),
    setuptools.Extension(
        "spokestack.extensions.webrtc.vad",
        ["spokestack/extensions/webrtc/vad.pyx"] + SOURCES,
        include_dirs=["filter_audio/agc/include/"],
    ),
]
with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="spokestack",
    version="0.0.16",
    author="Spokestack",
    author_email="support@spokestack.io",
    description="Spokestack Library for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/spokestack/spokestack-python",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    setup_requires=["setuptools", "numpy", "Cython"],
    install_requires=[
        "numpy",
        "Cython",
        "websocket_client",
        "tokenizers",
        "requests",
    ],
    ext_modules=cythonize(EXTENSIONS),
    include_dirs=[np.get_include()],
    zip_safe=False,
)



"""
TFLite model base class
"""
from typing import Any, List

import numpy as np

try:
    import tflite_runtime.interpreter as tflite
except ModuleNotFoundError:
    import tensorflow.lite as tflite


class TFLiteModel:
    """TFLite model base class for managing multiple inputs/outputs

    Args:
        model_path (str): Path to .tflite model file
        **kwargs (Any): Additional keywords arguments for the TFLite Interpreter.
                        [https://www.tensorflow.org/api_docs/python/tf/lite/Interpreter]
    """

    def __init__(self, model_path: str, **kwargs: Any) -> None:

        self._interpreter: tflite.Interpreter = tflite.Interpreter(
            model_path=model_path, **kwargs
        )
        self._input_details = self._interpreter.get_input_details()
        self._output_details = self._interpreter.get_output_details()
        self._interpreter.allocate_tensors()

    def __call__(self, *args: Any) -> List[np.ndarray]:
        """Forward pass of the TFLite model

        Args:
            inputs (Any): inputs to the TFLite model

        Returns: outputs of TFLite model

        """

        for detail, arg in zip(self._input_details, args):
            self._interpreter.set_tensor(detail["index"], arg)

        self._interpreter.invoke()

        return [
            self._interpreter.get_tensor(tensor["index"])
            for tensor in self._output_details
        ]

    @property
    def input_details(self) -> List[Any]:
        """Property for accesing the TFLite model input_details

        Returns: Input details for the TFLite model

        """
        return self._input_details

    @property
    def output_details(self) -> List[Any]:
        """Property for accesing the TFLite model output_details

        Returns: Output details for the TFLite model

        """
        return self._output_details



"""
Tests for TFLite model base class
"""
from unittest import mock

import numpy as np

from spokestack.models.tensorflow import TFLiteModel


@mock.patch("spokestack.models.tensorflow.tflite")
def test_inputs(_mock):
    model = TFLiteModel(model_path="model_path")
    model._input_details = [{"name": "inputs", "index": 0}]
    model._output_details = [{"name": "outputs", "index": 3}]

    # test single input
    one = np.zeros((1, 1))
    _ = model(one)

    # test multi input
    one = np.zeros((1, 1))
    _ = model(one, one)

    assert model.output_details
    assert model.input_details


def test_outputs():
    model = mock.MagicMock(return_value=[np.zeros((1, 1))])
    one = np.zeros((1, 1))
    outputs = model(one)
    assert len(outputs) == 1

    # test multi-output
    model = mock.MagicMock(return_value=[np.zeros((1, 1)), np.zeros((1, 1))])
    one = np.zeros((1, 1))
    outputs = model(one)
    assert len(outputs) > 1



"""
This module contains general purpose utilities for webrtc extensions
"""


class ProcessError(Exception):
    """Error for failure in processing"""

    def __init__(self, message: str):
        super().__init__(message)



"""
This module contains the tests for the spokestack text to speech manager
"""
from unittest import mock

import numpy as np

from spokestack.tts.manager import SequenceIO, TextToSpeechManager


def test_synthesize():
    client = mock.MagicMock()
    client.synthesize.return_value = np.zeros(100000, np.int16).tobytes()
    output = mock.MagicMock()

    with mock.patch("spokestack.tts.manager.MP3Decoder") as patched:
        patched.return_value = np.zeros(160).tobytes()
        manager = TextToSpeechManager(client, output)
        manager.synthesize(utterance="test utterance", mode="text", voice="demo-male")


def test_close():
    client = mock.MagicMock()
    output = mock.MagicMock()
    manager = TextToSpeechManager(client, output)
    manager.close()

    assert not manager._client
    assert not manager._output


def test_sequence_io():
    test = (np.ones(1000, np.int16).tobytes() for i in range(10))
    stream = SequenceIO(test)
    for _ in test:
        stream.read()
    # read after StopIteration
    stream.read()



"""
This module contains the logic to parse entities from NLU results. The entity parser
is a pass through for string values to allow custom logic to resolve the entities.
For example, the entity can be used as a keyword in a database search.
"""
from typing import Any, Dict


def parse(metadata: Dict[str, Any], raw_value: str) -> str:
    """Entity Parser

    Args:
        metadata (Dict[str, Any]): metadata for entity slot
        raw_value (str): tagged entity

    Returns:
        (str): tagged entity

    """
    return raw_value



"""
This module tests the cloud speech recognizer
"""
import json
from unittest import mock

import numpy as np

from spokestack.asr.spokestack.speech_recognizer import CloudSpeechRecognizer
from spokestack.context import SpeechContext


def test_recognize():
    context = SpeechContext()
    recognizer = CloudSpeechRecognizer()
    recognizer._client._socket = mock.MagicMock()

    recognizer._client._socket.recv.return_value = json.dumps(
        {
            "error": None,
            "final": False,
            "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
            "status": "ok",
        }
    )

    frame = np.random.rand(160).astype(np.int16)
    # call with context active to test _begin and first _send
    context.is_active = True
    recognizer(context, frame)

    # call again to test with internal _is_active as True
    recognizer(context, frame)

    # call with context not active to test _commit
    context.is_active = False
    recognizer(context, frame)

    recognizer._client._socket.recv.return_value = json.dumps(
        {
            "error": None,
            "final": True,
            "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
            "status": "ok",
        }
    )

    # call with the client indicating it's the final frame to test _receive
    recognizer(context, frame)

    recognizer._client._socket.max_idle_time = 500
    # test timeout
    for i in range(501):
        recognizer(context, frame)

    assert not context.is_active
    assert not recognizer._client.is_connected


def test_response():
    context = SpeechContext()
    recognizer = CloudSpeechRecognizer()
    recognizer._client._socket = mock.MagicMock()

    recognizer._client._socket.recv.return_value = json.dumps(
        {
            "error": None,
            "final": False,
            "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
            "status": "ok",
        }
    )

    frame = np.random.rand(160).astype(np.int16)

    # run through all the steps
    context.is_active = True
    recognizer(context, frame)
    recognizer(context, frame)
    context.is_active = False
    recognizer(context, frame)

    recognizer._client._socket.recv.return_value = json.dumps(
        {
            "error": None,
            "final": True,
            "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
            "status": "ok",
        }
    )
    # process the final frame with the final transcript
    recognizer(context, frame)

    assert context.transcript == "this is a test"
    assert context.confidence == 0.5

    recognizer.close()


def test_reset():
    context = SpeechContext()
    recognizer = CloudSpeechRecognizer()
    recognizer._client._socket = mock.MagicMock()

    recognizer._client._socket.recv.return_value = json.dumps(
        {
            "error": None,
            "final": False,
            "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
            "status": "ok",
        }
    )

    frame = np.random.rand(160).astype(np.int16)

    # trigger _begin and first _send
    context.is_active = True
    recognizer(context, frame)

    # trigger _send
    recognizer(context, frame)

    # we haven't triggered _commit or sent the final frame
    # which means context is still active and _is_active is True
    recognizer.reset()

    assert not recognizer._is_active
    assert not recognizer._client.is_connected


def test_empty_transcript():
    context = SpeechContext()
    recognizer = CloudSpeechRecognizer()
    recognizer._client._socket = mock.MagicMock()

    recognizer._client._socket.recv.return_value = json.dumps(
        {
            "error": None,
            "final": False,
            "hypotheses": [{"confidence": 0.5, "transcript": ""}],
            "status": "ok",
        }
    )

    frame = np.random.rand(160).astype(np.int16)

    # run through all the steps
    context.is_active = True
    recognizer(context, frame)
    recognizer(context, frame)
    context.is_active = False
    recognizer(context, frame)

    recognizer._client._socket.recv.return_value = json.dumps(
        {
            "error": None,
            "final": True,
            "hypotheses": [{"confidence": 0.5, "transcript": ""}],
            "status": "ok",
        }
    )
    # process the final frame with the final transcript
    recognizer(context, frame)

    assert not context.transcript
    assert context.confidence == 0.5

    recognizer.close()



"""
This module contains the speech pipeline which manages the components
for processing speech.
"""
from typing import Any, List, Union

from spokestack.context import SpeechContext


class SpeechPipeline:
    """Pipeline for managing speech components.

    Args:
        input_source: source of audio input
        stages: components desired in the pipeline
        **kwargs: additional keyword arguments
    """

    def __init__(self, input_source: Any, stages: List[Any]) -> None:
        self._context = SpeechContext()
        self._input_source = input_source
        self._stages: list = stages
        self._is_running = False
        self._is_paused = False

    def _dispatch(self) -> None:
        frame = self._input_source.read()
        for stage in self._stages:
            stage(self._context, frame)

    def close(self) -> None:
        """ Closes the running pipeline """
        self.stop()

    def activate(self) -> None:
        """ Activates the pipeline """
        self._context.is_active = True

    def deactivate(self) -> None:
        """ Deactivates the pipeline """
        self._context.is_active = False

    def start(self) -> None:
        """ Starts input source of the pipeline """
        if not self._is_running:
            self._input_source.start()
            self._is_running = True

    def stop(self) -> None:
        """ Halts the pipeline """
        self._is_running = False
        self._input_source.stop()

    def pause(self) -> None:
        """ Stops audio input until resume is called """
        self._is_paused = True
        self._input_source.stop()

    def resume(self) -> None:
        """ Resumes audio input after a pause """
        self._is_paused = False
        self._input_source.start()

    def run(self) -> None:
        """ Runs the pipeline to process speech and cleans up after stop is called """
        while self._is_running:
            self.step()
        self.cleanup()

    def step(self) -> None:
        """ Process a single frame with the pipeline """
        self._context.event("step")
        if not self._is_paused:
            self._dispatch()

    def cleanup(self) -> None:
        """ Resets all attributes to default configuration """
        for stage in self._stages:
            stage.close()

        self._stages.clear()
        self._input_source.close()
        self._input_source = None
        self._context.reset()

    def event(self, function: Any = None, name: Union[str, None] = None) -> Any:
        """Registers an event handler

        Args:
            function: event handler
            name: name of event handler

        Returns:
            Default event handler if a function not specified

        """
        if function:
            self._context.add_handler(
                name or function.__name__.replace("on_", ""), function
            )
        else:
            return lambda function: self.event(function, name)

    @property
    def is_running(self) -> bool:
        """ State of the pipeline """
        return self._is_running

    @property
    def context(self) -> SpeechContext:
        """ Current context """
        return self._context



"""
This module contains the logic to parse integers from NLU results. Integers can be
in the form of words (ie. one, two, three) or numbers (ie. 1, 2, 3). Either form
will resolve to Python's built-in 'int' type. The metadata must contain a range
key containing the minimum and maximum values for the expected integer range. It is
important to note the difference between digits and integers. Integers are
counting numbers: 2 apples, a table for two. In contrast, digits
can be used for sequences of numbers like phone numbers or social security numbers.
"""
from typing import Any, Dict, Union

from spokestack.nlu.parsers import DIGIT_SPLIT_RE, maps


def parse(metadata: Dict[str, Any], raw_value: str) -> Union[int, None]:
    """Integer Parser

    Args:
        metadata (Dict[str, Any]): metadata for the integer slot
        raw_value (str): value tagged by the model

    Returns:
        Union[int, None]: integer if parsable, None if invalid
    """
    raw_range = metadata.get("range")
    normalized = raw_value.lower()
    tokens = DIGIT_SPLIT_RE.split(normalized)

    parsed_values = []
    for token in tokens:
        try:
            parsed = int(token)
            parsed_values.append(parsed)
        except ValueError:
            if not _parse_reduce(token, parsed_values):
                return None

    result = sum(parsed_values)
    if _is_in_range(result, raw_range):
        return result
    return None


def _parse_reduce(number: Any, so_far: Any) -> Any:
    to_parse = number
    if to_parse.endswith("th"):
        to_parse = to_parse[: len(to_parse) - 2]
    if to_parse not in maps.WORD_TO_NUM:
        return None
    if to_parse in maps.MULTIPLIERS:
        total = _collapse(maps.MULTIPLIERS[to_parse], so_far)
        so_far.clear()
        so_far += total
    else:
        so_far.append(maps.WORD_TO_NUM[to_parse])
    return so_far


def _collapse(multiplier: int, so_far: Any) -> Any:
    collapsed = []
    total = 0
    for number in so_far:
        if number > multiplier:
            collapsed.append(number)
        else:
            total += number
    total = max(total, 1)
    collapsed.append(total * multiplier)
    return collapsed


def _is_in_range(value: int, interval: Any) -> bool:
    return value in range(interval[0], interval[1])



"""
Pipeline profile for pyaudio input, vad, wakeword, and asr
"""
from typing import Any

from spokestack.activation_timeout import ActivationTimeout
from spokestack.agc.webrtc import AutomaticGainControl
from spokestack.asr.spokestack.speech_recognizer import CloudSpeechRecognizer
from spokestack.io.pyaudio import PyAudioInput
from spokestack.nsx.webrtc import AutomaticNoiseSuppression
from spokestack.pipeline import SpeechPipeline
from spokestack.vad.webrtc import VoiceActivityDetector
from spokestack.wakeword.tflite import WakewordTrigger


class WakewordSpokestackASR:
    """ TFLite wakeword with Spokestack speech recognition. """

    @staticmethod
    def create(
        spokestack_id: str,
        spokestack_secret: str,
        sample_rate: int = 16000,
        frame_width: int = 20,
        model_dir: str = "",
        **kwargs: Any,
    ) -> SpeechPipeline:
        """Creates a speech pipeline instance from profile

        Args:
            spokestack_id (str): spokestack API id.
            spokestack_secret (str): Spokestack API secret.
            sample_rate (int): sample rate of the audio (Hz).
            frame_width (int): width of the audio frame: 10, 20, or 30 (ms).
            model_dir (str): Directory containing the tflite wakeword models.

        Returns:

        """
        pipeline = SpeechPipeline(
            input_source=PyAudioInput(
                frame_width=frame_width, sample_rate=sample_rate, **kwargs
            ),
            stages=[
                AutomaticGainControl(sample_rate=sample_rate, frame_width=frame_width),
                AutomaticNoiseSuppression(sample_rate=sample_rate),
                VoiceActivityDetector(
                    frame_width=frame_width,
                    sample_rate=sample_rate,
                    **kwargs,
                ),
                WakewordTrigger(model_dir=model_dir, **kwargs),
                ActivationTimeout(frame_width=frame_width, **kwargs),
                CloudSpeechRecognizer(
                    spokestack_secret=spokestack_secret,
                    spokestack_id=spokestack_id,
                    **kwargs,
                ),
            ],
        )
        return pipeline



"""
This module contains the tests for the cloud-based asr client
"""
import json
from unittest import mock

import numpy as np
import pytest

from spokestack.asr.spokestack.cloud_client import APIError, CloudClient


@mock.patch("spokestack.asr.spokestack.cloud_client.WebSocket")
def test_socket_connect(_mock):
    client = CloudClient(socket_url="", key_id="", key_secret="")

    with pytest.raises(ConnectionError):
        client.initialize()

    assert not client.is_connected
    client.connect()

    client._socket.recv.return_value = json.dumps(
        {
            "error": None,
            "final": False,
            "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
            "status": "ok",
        }
    )
    assert client.is_connected

    # try a double connect which should be silently ignored
    client.connect()

    client.disconnect()
    assert not client.is_connected


@mock.patch("spokestack.asr.spokestack.cloud_client.WebSocket")
def test_send_audio(_mock):
    client = CloudClient(socket_url="", key_id="", key_secret="")
    frame = np.random.rand(160).astype(np.int16)

    with pytest.raises(ConnectionError):
        client.send(frame)

    with pytest.raises(ConnectionError):
        client.end()

    client.connect()
    client.end()

    for i in range(100):
        client.send(frame)
    client.end()


@mock.patch("spokestack.asr.spokestack.cloud_client.WebSocket")
def test_reponse_event(_mock):
    client = CloudClient(socket_url="", key_id="", key_secret="")

    with pytest.raises(ConnectionError):
        client.receive()

    client.connect()
    client._socket.recv.return_value = json.dumps(
        {"error": None, "final": False, "hypotheses": [], "status": "ok"}
    )
    client.initialize()
    client._socket.recv.return_value = json.dumps(
        {
            "error": None,
            "final": False,
            "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
            "status": "ok",
        }
    )

    client.receive()
    assert not client.response["error"]
    assert not client.response["final"]
    assert len(client.response["hypotheses"]) > 0
    assert client.response["status"] == "ok"

    client._socket.recv.return_value = json.dumps(
        {
            "error": None,
            "final": False,
            "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
            "status": "ok",
        }
    )

    client.receive()
    assert not client.response["error"]
    assert not client.response["final"]
    assert client.response["hypotheses"][0]["confidence"] == 0.5
    assert client.response["hypotheses"][0]["transcript"] == "this is a test"
    assert client.response["status"] == "ok"

    client._socket.recv.return_value = json.dumps(
        {
            "error": None,
            "final": True,
            "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
            "status": "ok",
        }
    )

    client.receive()
    assert client.is_final


@mock.patch("spokestack.asr.spokestack.cloud_client.WebSocket")
def test_call(_mock):
    client = CloudClient(socket_url="", key_id="", key_secret="", idle_timeout=5000)
    client.connect()
    client._socket.recv.side_effect = [
        json.dumps(
            {
                "error": None,
                "final": False,
                "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
                "status": "ok",
            }
        ),
        json.dumps(
            {
                "error": None,
                "final": False,
                "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
                "status": "ok",
            }
        ),
        json.dumps(
            {
                "error": None,
                "final": False,
                "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
                "status": "ok",
            }
        ),
        json.dumps(
            {
                "error": None,
                "final": True,
                "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
                "status": "ok",
            }
        ),
    ]
    client.initialize()
    audio = np.random.rand(160 * 50).astype(np.int16)

    transcript = client(audio)[0]
    assert transcript == {"confidence": 0.5, "transcript": "this is a test"}


@mock.patch("spokestack.asr.spokestack.cloud_client.WebSocket")
def test_api_error(_mock):
    client = CloudClient(socket_url="", key_id="", key_secret="", idle_timeout=5000)

    client.connect()
    client._socket.recv.return_value = json.dumps(
        {
            "error": "invalid_language",
            "final": False,
            "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
            "status": "error",
        }
    )
    with pytest.raises(APIError):
        client.initialize()


@mock.patch("spokestack.asr.spokestack.cloud_client.WebSocket")
def test_bad_recv(_mock):
    client = CloudClient(socket_url="", key_id="", key_secret="", idle_timeout=5000)
    client.connect()
    client._socket.recv.side_effect = [
        json.dumps(
            {
                "error": None,
                "final": False,
                "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
                "status": "ok",
            }
        ),
        # bad response
        {},
    ]
    client.initialize()
    client.receive()


@mock.patch("spokestack.asr.spokestack.cloud_client.WebSocket")
def test_type_conversions(_mock):
    dummpy_inputs = [
        json.dumps(
            {
                "error": None,
                "final": False,
                "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
                "status": "ok",
            }
        ),
        json.dumps(
            {
                "error": None,
                "final": False,
                "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
                "status": "ok",
            }
        ),
        json.dumps(
            {
                "error": None,
                "final": True,
                "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
                "status": "ok",
            }
        ),
    ]
    client = CloudClient(socket_url="", key_id="", key_secret="")
    client.connect()
    client._socket.recv.side_effect = dummpy_inputs

    client.initialize()
    audio = np.zeros((100), np.float32)

    # as np.float32
    transcript = client(audio)[0]
    assert transcript == {"confidence": 0.5, "transcript": "this is a test"}

    # as bytes
    client.connect()
    client._socket.recv.side_effect = dummpy_inputs
    client.initialize()
    transcript = client(audio.tobytes())[0]
    assert transcript == {"confidence": 0.5, "transcript": "this is a test"}

    # as int16
    client.connect()
    client._socket.recv.side_effect = dummpy_inputs
    client.initialize()
    transcript = client(audio.astype(np.int16))[0]
    assert transcript == {"confidence": 0.5, "transcript": "this is a test"}

    # as float64
    client.connect()
    client._socket.recv.side_effect = dummpy_inputs
    client.initialize()
    transcript = client(audio.astype(np.float64))[0]
    assert transcript == {"confidence": 0.5, "transcript": "this is a test"}

    # as invalid
    client.connect()
    client._socket.recv.side_effect = dummpy_inputs
    client.initialize()
    with pytest.raises(TypeError):
        _ = client(audio.astype(np.complex64))[0]


@mock.patch("spokestack.asr.spokestack.cloud_client.WebSocket")
def test_is_connected(_mock):
    client = CloudClient(socket_url="", key_id="", key_secret="")
    client.connect()
    assert client.is_connected
    client.disconnect()
    assert not client.is_connected


@mock.patch("spokestack.asr.spokestack.cloud_client.WebSocket")
def test_initialize(_mock):
    client = CloudClient(socket_url="", key_id="", key_secret="")
    client.connect()
    assert client.is_connected
    client._socket.recv.return_value = json.dumps(
        {"error": None, "final": False, "hypotheses": [], "status": "ok"}
    )
    client.initialize()


@mock.patch("spokestack.asr.spokestack.cloud_client.WebSocket")
def test_disconnect(_mock):
    client = CloudClient(socket_url="", key_id="", key_secret="")

    assert not client.is_connected
    client.connect()
    assert client.is_connected
    client.disconnect()
    assert not client.is_connected


@mock.patch("spokestack.asr.spokestack.cloud_client.WebSocket")
def test_end(_mock):
    client = CloudClient(socket_url="", key_id="", key_secret="")
    client.connect()
    client._socket.recv.side_effect = [
        json.dumps(
            {
                "error": None,
                "final": False,
                "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
                "status": "ok",
            }
        ),
        json.dumps(
            {
                "error": None,
                "final": True,
                "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
                "status": "ok",
            }
        ),
    ]

    client.initialize()
    client.end()
    client.receive()
    assert client.is_final


@mock.patch("spokestack.asr.spokestack.cloud_client.WebSocket")
def test_response(_mock):
    client = CloudClient(socket_url="", key_id="", key_secret="")
    response = {
        "error": None,
        "final": False,
        "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
        "status": "ok",
    }
    client.connect()
    client._socket.recv.return_value = json.dumps(response)
    client.receive()
    assert client.response == response


@mock.patch("spokestack.asr.spokestack.cloud_client.WebSocket")
def test_is_final(_mock):
    client = CloudClient(socket_url="", key_id="", key_secret="")
    client.connect()
    client._socket.recv.return_value = json.dumps(
        {
            "error": None,
            "final": True,
            "hypotheses": [{"confidence": 0.5, "transcript": "this is a test"}],
            "status": "ok",
        }
    )
    client.receive()
    assert client.is_final


@mock.patch("spokestack.asr.spokestack.cloud_client.WebSocket")
def test_idle_timeout(_mock):
    client = CloudClient(socket_url="", key_id="", key_secret="", idle_timeout=500)
    assert client.idle_timeout == 500


@mock.patch("spokestack.asr.spokestack.cloud_client.WebSocket")
def test_idle_count(_mock):
    client = CloudClient(socket_url="", key_id="", key_secret="")
    assert client.idle_count == 0

    for i in range(5):
        client.idle_count += 1

    assert client.idle_count == 5



cdef extern from "filter_audio/vad/include/webrtc_vad.h":
    ctypedef struct VadInst:
        pass

    int WebRtcVad_Create(VadInst ** handle)
    void WebRtcVad_Free(VadInst*handle)
    int WebRtcVad_Init(VadInst*handle)
    int WebRtcVad_set_mode(VadInst*handle, int mode)
    bint WebRtcVad_Process(VadInst*handle, int fs, short*audio_frame, int frame_length)
